# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from watcherclient.v1 import action
from watcherclient.v1 import action_plan
from watcherclient.v1 import audit
from watcherclient.v1 import audit_template
from watcherclient.v1 import data_model
from watcherclient.v1 import goal
from watcherclient.v1 import scoring_engine
from watcherclient.v1 import service
from watcherclient.v1 import strategy

Action = action.Action
ActionManager = action.ActionManager
ActionPlan = action_plan.ActionPlan
ActionPlanManager = action_plan.ActionPlanManager
Audit = audit.Audit
AuditManager = audit.AuditManager
AuditTemplate = audit_template.AuditTemplate
AuditTemplateManager = audit_template.AuditTemplateManager
Goal = goal.Goal
GoalManager = goal.GoalManager
ScoringEngine = scoring_engine.ScoringEngine
ScoringEngineManager = scoring_engine.ScoringEngineManager
Service = service.Service
ServiceManager = service.ServiceManager
Strategy = strategy.Strategy
StrategyManager = strategy.StrategyManager
DataModel = data_model.DataModel
DataModelManager = data_model.DataModelManager

__all__ = (
    "Action", "ActionManager", "ActionPlan", "ActionPlanManager",
    "Audit", "AuditManager", "AuditTemplate", "AuditTemplateManager",
    "Goal", "GoalManager", "ScoringEngine", "ScoringEngineManager",
    "Service", "ServiceManager", "Strategy", "StrategyManager",
    "DataModel", "DataModelManager")



# Copyright (c) 2013 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Command-line interface to the Watcher API.
"""
from collections import namedtuple
import logging
import sys

from cliff import app
from cliff import command
from cliff import commandmanager
from cliff import complete
from cliff import help as cli_help
from keystoneauth1 import loading
from osc_lib import logs
from osc_lib import utils

from watcherclient import client as watcherclient
from watcherclient import version

LOG = logging.getLogger(__name__)


API_NAME = 'infra-optim'
API_VERSIONS = {
    '1': 'watcherclient.v1.client.Client',
}
DEFAULT_OS_INFRA_OPTIM_API_VERSION = '1.latest'
_DEFAULT_IDENTITY_API_VERSION = '3'
_IDENTITY_API_VERSION_2 = ['2', '2.0']
_IDENTITY_API_VERSION_3 = ['3']


class WatcherShell(app.App):
    """Watcher command line interface."""

    def __init__(self, **kwargs):
        self.client = None

        # Patch command.Command to add a default auth_required = True
        command.Command.auth_required = True

        # Some commands do not need authentication
        cli_help.HelpCommand.auth_required = False
        complete.CompleteCommand.auth_required = False

        super(WatcherShell, self).__init__(
            description=__doc__.strip(),
            version=version.__version__,
            command_manager=commandmanager.CommandManager(
                'watcherclient.v1'),
            deferred_help=True,
            **kwargs
        )

    def create_client(self, args):
        client = watcherclient.get_client('1', **args.__dict__)
        return client

    def build_option_parser(self, description, version, argparse_kwargs=None):
        """Introduces global arguments for the application.

        This is inherited from the framework.
        """
        parser = super(WatcherShell, self).build_option_parser(
            description, version, argparse_kwargs)
        parser.add_argument('--no-auth', '-N', action='store_true',
                            help='Do not use authentication.')
        parser.add_argument('--os-identity-api-version',
                            metavar='<identity-api-version>',
                            default=utils.env('OS_IDENTITY_API_VERSION'),
                            help='Specify Identity API version to use. '
                            'Defaults to env[OS_IDENTITY_API_VERSION]'
                            ' or 3.')
        parser.add_argument('--os-auth-url', '-A',
                            metavar='<auth-url>',
                            default=utils.env('OS_AUTH_URL'),
                            help='Defaults to env[OS_AUTH_URL].')
        parser.add_argument('--os-region-name', '-R',
                            metavar='<region-name>',
                            default=utils.env('OS_REGION_NAME'),
                            help='Defaults to env[OS_REGION_NAME].')
        parser.add_argument('--os-username', '-U',
                            metavar='<auth-user-name>',
                            default=utils.env('OS_USERNAME'),
                            help='Defaults to env[OS_USERNAME].')
        parser.add_argument('--os-user-id',
                            metavar='<auth-user-id>',
                            default=utils.env('OS_USER_ID'),
                            help='Defaults to env[OS_USER_ID].')
        parser.add_argument('--os-password', '-P',
                            metavar='<auth-password>',
                            default=utils.env('OS_PASSWORD'),
                            help='Defaults to env[OS_PASSWORD].')
        parser.add_argument('--os-user-domain-id',
                            metavar='<auth-user-domain-id>',
                            default=utils.env('OS_USER_DOMAIN_ID'),
                            help='Defaults to env[OS_USER_DOMAIN_ID].')
        parser.add_argument('--os-user-domain-name',
                            metavar='<auth-user-domain-name>',
                            default=utils.env('OS_USER_DOMAIN_NAME'),
                            help='Defaults to env[OS_USER_DOMAIN_NAME].')
        parser.add_argument('--os-tenant-name', '-T',
                            metavar='<auth-tenant-name>',
                            default=utils.env('OS_TENANT_NAME'),
                            help='Defaults to env[OS_TENANT_NAME].')
        parser.add_argument('--os-tenant-id', '-I',
                            metavar='<tenant-id>',
                            default=utils.env('OS_TENANT_ID'),
                            help='Defaults to env[OS_TENANT_ID].')
        parser.add_argument('--os-project-id',
                            metavar='<auth-project-id>',
                            default=utils.env('OS_PROJECT_ID'),
                            help='Another way to specify tenant ID. '
                                 'This option is mutually exclusive with '
                                 ' --os-tenant-id. '
                            'Defaults to env[OS_PROJECT_ID].')
        parser.add_argument('--os-project-name',
                            metavar='<auth-project-name>',
                            default=utils.env('OS_PROJECT_NAME'),
                            help='Another way to specify tenant name. '
                                 'This option is mutually exclusive with '
                                 ' --os-tenant-name. '
                                 'Defaults to env[OS_PROJECT_NAME].')
        parser.add_argument('--os-project-domain-id',
                            metavar='<auth-project-domain-id>',
                            default=utils.env('OS_PROJECT_DOMAIN_ID'),
                            help='Defaults to env[OS_PROJECT_DOMAIN_ID].')
        parser.add_argument('--os-project-domain-name',
                            metavar='<auth-project-domain-name>',
                            default=utils.env('OS_PROJECT_DOMAIN_NAME'),
                            help='Defaults to env[OS_PROJECT_DOMAIN_NAME].')
        parser.add_argument('--os-auth-token',
                            metavar='<auth-token>',
                            default=utils.env('OS_AUTH_TOKEN'),
                            help='Defaults to env[OS_AUTH_TOKEN].')
        parser.add_argument(
            '--os-infra-optim-api-version',
            metavar='<infra-optim-api-version>',
            default=utils.env('OS_INFRA_OPTIM_API_VERSION',
                              default=DEFAULT_OS_INFRA_OPTIM_API_VERSION),
            help='Accepts X, X.Y (where X is major and Y is minor part) or '
                 '"X.latest", defaults to env[OS_INFRA_OPTIM_API_VERSION].')
        parser.add_argument('--os-endpoint-type',
                            default=utils.env('OS_ENDPOINT_TYPE'),
                            help='Defaults to env[OS_ENDPOINT_TYPE] or '
                                 '"publicURL"')
        parser.add_argument('--os-endpoint-override',
                            metavar='<endpoint-override>',
                            default=utils.env('OS_ENDPOINT_OVERRIDE'),
                            help="Use this API endpoint instead of the "
                                 "Service Catalog.")
        parser.epilog = ('See "watcher help COMMAND" for help '
                         'on a specific command.')
        loading.register_session_argparse_arguments(parser)
        return parser

    def configure_logging(self):
        """Configure logging for the app."""
        self.log_configurator = logs.LogConfigurator(self.options)
        self.dump_stack_trace = self.log_configurator.dump_trace

    def prepare_to_run_command(self, cmd):
        """Prepares to run the command

        Checks if the minimal parameters are provided and creates the
        client interface.
        This is inherited from the framework.
        """
        self.client_manager = namedtuple('ClientManager', 'infra_optim')
        if cmd.auth_required:
            client = self.create_client(self.options)
            setattr(self.client_manager, 'infra-optim', client)

    def run(self, argv):
        ret_val = 1
        self.command_options = argv
        try:
            ret_val = super(WatcherShell, self).run(argv)
            return ret_val
        except Exception as e:
            if not logging.getLogger('').handlers:
                logging.basicConfig()
            LOG.error('Exception raised: %s', str(e))

            return ret_val

        finally:
            LOG.info("END return value: %s", ret_val)


def main(argv=sys.argv[1:]):
    watcher_app = WatcherShell()
    return watcher_app.run(argv)


if __name__ == '__main__':   # pragma: no cover
    sys.exit(main(sys.argv[1:]))



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from distutils import version
import functools
import hashlib
import logging
import os
import socket
import ssl
import textwrap
import time

from keystoneauth1 import adapter
from keystoneauth1 import exceptions as kexceptions
from oslo_serialization import jsonutils
from oslo_utils import strutils
import requests
import six
from six.moves import http_client
import six.moves.urllib.parse as urlparse

from watcherclient._i18n import _
from watcherclient.common import api_versioning
from watcherclient import exceptions


# Record the latest version that this client was tested with.
DEFAULT_VER = '1.latest'
# Minor version 4 for adding webhook API
LAST_KNOWN_API_VERSION = 4
LATEST_VERSION = '1.{}'.format(LAST_KNOWN_API_VERSION)

LOG = logging.getLogger(__name__)
USER_AGENT = 'python-watcherclient'
CHUNKSIZE = 1024 * 64  # 64kB

API_VERSION = '/v1'
API_VERSION_SELECTED_STATES = ('user', 'negotiated', 'cached', 'default')


DEFAULT_MAX_RETRIES = 5
DEFAULT_RETRY_INTERVAL = 2
SENSITIVE_HEADERS = ('X-Auth-Token',)


SUPPORTED_ENDPOINT_SCHEME = ('http', 'https')


def _trim_endpoint_api_version(url):
    """Trim API version and trailing slash from endpoint."""
    return url.rstrip('/').rstrip(API_VERSION)


def _extract_error_json(body):
    """Return  error_message from the HTTP response body."""
    error_json = {}
    try:
        body_json = jsonutils.loads(body)
        if 'error_message' in body_json:
            raw_msg = body_json['error_message']
            error_json = jsonutils.loads(raw_msg)
    except ValueError:
        pass

    return error_json


def get_server(endpoint):
    """Extract and return the server & port that we're connecting to."""
    if endpoint is None:
        return None, None
    parts = urlparse.urlparse(endpoint)
    return parts.hostname, str(parts.port)


class VersionNegotiationMixin(object):
    def negotiate_version(self, conn, resp):
        """Negotiate the server version

        Assumption: Called after receiving a 406 error when doing a request.

        param conn: A connection object
        param resp: The response object from http request
        """
        if self.api_version_select_state not in API_VERSION_SELECTED_STATES:
            raise RuntimeError(
                _('Error: self.api_version_select_state should be one of the '
                  'values in: "%(valid)s" but had the value: "%(value)s"') %
                {'valid': ', '.join(API_VERSION_SELECTED_STATES),
                 'value': self.api_version_select_state})
        min_ver, max_ver = self._parse_version_headers(resp)
        # If the user requested an explicit version or we have negotiated a
        # version and still failing then error now. The server could
        # support the version requested but the requested operation may not
        # be supported by the requested version.
        if self.api_version_select_state == 'user':
            raise exceptions.UnsupportedVersion(textwrap.fill(
                _("Requested API version %(req)s is not supported by the "
                  "server or the requested operation is not supported by the "
                  "requested version.  Supported version range is %(min)s to "
                  "%(max)s")
                % {'req': self.os_infra_optim_api_version,
                   'min': min_ver, 'max': max_ver}))
        if self.api_version_select_state == 'negotiated':
            raise exceptions.UnsupportedVersion(textwrap.fill(
                _("No API version was specified and the requested operation "
                  "was not supported by the client's negotiated API version "
                  "%(req)s.  Supported version range is: %(min)s to %(max)s")
                % {'req': self.os_infra_optim_api_version,
                   'min': min_ver, 'max': max_ver}))

        negotiated_ver = str(
            min(version.StrictVersion(self.os_infra_optim_api_version),
                version.StrictVersion(max_ver)))
        if negotiated_ver < min_ver:
            negotiated_ver = min_ver
        # server handles microversions, but doesn't support
        # the requested version, so try a negotiated version
        self.api_version_select_state = 'negotiated'
        self.os_infra_optim_api_version = negotiated_ver
        LOG.debug('Negotiated API version is %s', negotiated_ver)

        return negotiated_ver

    def _generic_parse_version_headers(self, accessor_func):
        min_ver = accessor_func('OpenStack-API-Minimum-Version',
                                None)
        max_ver = accessor_func('OpenStack-API-Maximum-Version',
                                None)
        return min_ver, max_ver

    def _parse_version_headers(self, accessor_func):
        # NOTE(jlvillal): Declared for unit testing purposes
        raise NotImplementedError()

    def _make_simple_request(self, conn, method, url):
        # NOTE(jlvillal): Declared for unit testing purposes
        raise NotImplementedError()


_RETRY_EXCEPTIONS = (exceptions.ServiceUnavailable,
                     exceptions.ConnectionRefused,
                     kexceptions.RetriableConnectionFailure)


def with_retries(func):
    """Wrapper for _http_request adding support for retries."""
    @functools.wraps(func)
    def wrapper(self, url, method, **kwargs):
        if self.conflict_max_retries is None:
            self.conflict_max_retries = DEFAULT_MAX_RETRIES
        if self.conflict_retry_interval is None:
            self.conflict_retry_interval = DEFAULT_RETRY_INTERVAL

        num_attempts = self.conflict_max_retries + 1
        for attempt in range(1, num_attempts + 1):
            try:
                return func(self, url, method, **kwargs)
            except _RETRY_EXCEPTIONS as error:
                msg = ("Error contacting Watcher server: %(error)s. "
                       "Attempt %(attempt)d of %(total)d" %
                       {'attempt': attempt,
                        'total': num_attempts,
                        'error': error})
                if attempt == num_attempts:
                    LOG.error(msg)
                    raise
                else:
                    LOG.debug(msg)
                    time.sleep(self.conflict_retry_interval)

    return wrapper


class HTTPClient(VersionNegotiationMixin):

    def __init__(self, endpoint, **kwargs):
        self.endpoint = endpoint
        self.endpoint_trimmed = _trim_endpoint_api_version(endpoint)
        self.auth_token = kwargs.get('token')
        self.auth_ref = kwargs.get('auth_ref')
        self.os_infra_optim_api_version = kwargs.get(
            'os_infra_optim_api_version', DEFAULT_VER)
        self.api_version_select_state = kwargs.get(
            'api_version_select_state', 'default')
        self.conflict_max_retries = kwargs.pop('max_retries',
                                               DEFAULT_MAX_RETRIES)
        self.conflict_retry_interval = kwargs.pop('retry_interval',
                                                  DEFAULT_RETRY_INTERVAL)
        self.session = requests.Session()

        parts = urlparse.urlparse(endpoint)
        if parts.scheme not in SUPPORTED_ENDPOINT_SCHEME:
            msg = _('Unsupported scheme: %s') % parts.scheme
            raise exceptions.EndpointException(msg)

        if parts.scheme == 'https':
            if kwargs.get('insecure') is True:
                self.session.verify = False
            elif kwargs.get('ca_file'):
                self.session.verify = kwargs['ca_file']
            self.session.cert = (kwargs.get('cert_file'),
                                 kwargs.get('key_file'))

    def _process_header(self, name, value):
        """Redacts any sensitive header

        Redact a header that contains sensitive information, by returning an
        updated header with the sha1 hash of that value. The redacted value is
        prefixed by '{SHA1}' because that's the convention used within
        OpenStack.

        :returns: A tuple of (name, value)
                  name: the safe encoding format of name
                  value: the redacted value if name is x-auth-token,
                         or the safe encoding format of name

        """
        if name in SENSITIVE_HEADERS:
            v = value.encode('utf-8')
            h = hashlib.sha1(v)
            d = h.hexdigest()
            return (name, "{SHA1}%s" % d)
        else:
            return (name, value)

    def log_curl_request(self, method, url, kwargs):
        curl = ['curl -i -X %s' % method]

        for (key, value) in kwargs['headers'].items():
            header = '-H \'%s: %s\'' % self._process_header(key, value)
            curl.append(header)

        if not self.session.verify:
            curl.append('-k')
        elif isinstance(self.session.verify, six.string_types):
            curl.append('--cacert %s' % self.session.verify)

        if self.session.cert:
            curl.append('--cert %s' % self.session.cert[0])
            curl.append('--key %s' % self.session.cert[1])

        if 'body' in kwargs:
            body = strutils.mask_password(kwargs['body'])
            curl.append('-d \'%s\'' % body)

        curl.append(urlparse.urljoin(self.endpoint_trimmed, url))
        LOG.debug(' '.join(curl))

    @staticmethod
    def log_http_response(resp, body=None):
        # NOTE(aarefiev): resp.raw is urllib3 response object, it's used
        # only to get 'version', response from request with 'stream = True'
        # should be used for raw reading.
        status = (resp.raw.version / 10.0, resp.status_code, resp.reason)
        dump = ['\nHTTP/%.1f %s %s' % status]
        dump.extend(['%s: %s' % (k, v) for k, v in resp.headers.items()])
        dump.append('')
        if body:
            body = strutils.mask_password(body)
            dump.extend([body, ''])
        LOG.debug('\n'.join(dump))

    def _make_connection_url(self, url):
        return '%s/%s' % (self.endpoint_trimmed.rstrip('/'), url.lstrip('/'))

    def _parse_version_headers(self, resp):
        return self._generic_parse_version_headers(resp.headers.get)

    def _make_simple_request(self, conn, method, url):
        return conn.request(method, self._make_connection_url(url))

    @with_retries
    def _http_request(self, url, method, **kwargs):
        """Send an http request with the specified characteristics.

        Wrapper around request.Session.request to handle tasks such
        as setting headers and error handling.
        """
        # Copy the kwargs so we can reuse the original in case of redirects
        kwargs['headers'] = copy.deepcopy(kwargs.get('headers', {}))
        kwargs['headers'].setdefault('User-Agent', USER_AGENT)
        if self.os_infra_optim_api_version:
            api_version = api_versioning.get_api_version(
                self.os_infra_optim_api_version)
            if api_version.is_latest():
                api_version = api_versioning.get_api_version(
                    LATEST_VERSION)
            kwargs['headers'].setdefault(
                'OpenStack-API-Version',
                ' '.join(['infra-optim', api_version.get_string()]))
        if self.auth_token:
            kwargs['headers'].setdefault('X-Auth-Token', self.auth_token)

        self.log_curl_request(method, url, kwargs)

        # NOTE(aarefiev): This is for backwards compatibility, request
        # expected body in 'data' field, previously we used httplib,
        # which expected 'body' field.
        body = kwargs.pop('body', None)
        if body:
            kwargs['data'] = body

        conn_url = self._make_connection_url(url)
        try:
            resp = self.session.request(method,
                                        conn_url,
                                        **kwargs)

            # TODO(deva): implement graceful client downgrade when connecting
            # to servers that did not support microversions. Details here:
            # http://specs.openstack.org/openstack/watcher-specs/specs/kilo/api-microversions.html#use-case-3b-new-client-communicating-with-a-old-watcher-user-specified  # noqa

            if resp.status_code == http_client.NOT_ACCEPTABLE:
                negotiated_ver = self.negotiate_version(self.session, resp)
                kwargs['headers']['OpenStack-API-Version'] = (
                    ' '.join(['infra-optim', negotiated_ver]))
                return self._http_request(url, method, **kwargs)

        except requests.exceptions.RequestException as e:
            message = (_("Error has occurred while handling "
                       "request for %(url)s: %(e)s") %
                       dict(url=conn_url, e=e))
            # NOTE(aarefiev): not valid request(invalid url, missing schema,
            # and so on), retrying is not needed.
            if isinstance(e, ValueError):
                raise exceptions.ValidationError(message)

            raise exceptions.ConnectionRefused(message)

        body_iter = resp.iter_content(chunk_size=CHUNKSIZE)

        # Read body into string if it isn't obviously image data
        body_str = None
        if resp.headers.get('Content-Type') != 'application/octet-stream':
            # decoding byte to string is necessary for Python 3 compatibility
            # this issues has not been found with Python 3 unit tests
            # because the test creates a fake http response of type str
            # the if statement satisfies test (str) and real (bytes) behavior
            body_list = [
                chunk.decode("utf-8") if isinstance(chunk, bytes)
                else chunk for chunk in body_iter
            ]
            body_str = ''.join(body_list)
            self.log_http_response(resp, body_str)
            body_iter = six.StringIO(body_str)
        else:
            self.log_http_response(resp)

        if resp.status_code >= http_client.BAD_REQUEST:
            error_json = _extract_error_json(body_str)
            raise exceptions.from_response(
                resp, error_json.get('faultstring'),
                error_json.get('debuginfo'), method, url)
        elif resp.status_code in (http_client.MOVED_PERMANENTLY,
                                  http_client.FOUND,
                                  http_client.USE_PROXY):
            # Redirected. Reissue the request to the new location.
            return self._http_request(resp['location'], method, **kwargs)
        elif resp.status_code == http_client.MULTIPLE_CHOICES:
            raise exceptions.from_response(resp, method=method, url=url)

        return resp, body_iter

    def json_request(self, method, url, **kwargs):
        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Content-Type', 'application/json')
        kwargs['headers'].setdefault('Accept', 'application/json')

        if 'body' in kwargs:
            kwargs['body'] = jsonutils.dumps(kwargs['body'])

        resp, body_iter = self._http_request(url, method, **kwargs)
        content_type = resp.headers.get('Content-Type')

        if (resp.status_code in (http_client.NO_CONTENT,
                                 http_client.RESET_CONTENT) or
                content_type is None):
            return resp, list()

        if 'application/json' in content_type:
            body = ''.join([chunk for chunk in body_iter])
            try:
                body = jsonutils.loads(body)
            except ValueError:
                LOG.error('Could not decode response body as JSON')
        else:
            body = None

        return resp, body

    def raw_request(self, method, url, **kwargs):
        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Content-Type',
                                     'application/octet-stream')
        return self._http_request(url, method, **kwargs)


class VerifiedHTTPSConnection(six.moves.http_client.HTTPSConnection):
    """httplib-compatible connection using client-side SSL authentication

    :see http://code.activestate.com/recipes/
            577548-https-httplib-client-connection-with-certificate-v/
    """

    def __init__(self, host, port, key_file=None, cert_file=None,
                 ca_file=None, timeout=None, insecure=False):
        six.moves.http_client.HTTPSConnection.__init__(self, host, port,
                                                       key_file=key_file,
                                                       cert_file=cert_file)
        self.key_file = key_file
        self.cert_file = cert_file
        if ca_file is not None:
            self.ca_file = ca_file
        else:
            self.ca_file = self.get_system_ca_file()
        self.timeout = timeout
        self.insecure = insecure

    def connect(self):
        """Connect to a host on a given (SSL) port.

        If ca_file is pointing somewhere, use it to check Server Certificate.

        Redefined/copied and extended from httplib.py:1105 (Python 2.6.x).
        This is needed to pass cert_reqs=ssl.CERT_REQUIRED as parameter to
        ssl.wrap_socket(), which forces SSL to check server certificate against
        our client certificate.
        """
        sock = socket.create_connection((self.host, self.port), self.timeout)

        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        if self.insecure is True:
            kwargs = {'cert_reqs': ssl.CERT_NONE}
        else:
            kwargs = {'cert_reqs': ssl.CERT_REQUIRED, 'ca_certs': self.ca_file}

        if self.cert_file:
            kwargs['certfile'] = self.cert_file
            if self.key_file:
                kwargs['keyfile'] = self.key_file

        self.sock = ssl.wrap_socket(sock, **kwargs)

    @staticmethod
    def get_system_ca_file():
        """Return path to system default CA file."""
        # Standard CA file locations for Debian/Ubuntu, RedHat/Fedora,
        # Suse, FreeBSD/OpenBSD
        ca_path = ['/etc/ssl/certs/ca-certificates.crt',
                   '/etc/pki/tls/certs/ca-bundle.crt',
                   '/etc/ssl/ca-bundle.pem',
                   '/etc/ssl/cert.pem']
        for ca in ca_path:
            if os.path.exists(ca):
                return ca
        return None


class SessionClient(VersionNegotiationMixin, adapter.LegacyJsonAdapter):
    """HTTP client based on Keystone client session."""

    def __init__(self,
                 os_infra_optim_api_version,
                 api_version_select_state,
                 max_retries,
                 retry_interval,
                 endpoint,
                 **kwargs):
        self.os_infra_optim_api_version = os_infra_optim_api_version
        self.api_version_select_state = api_version_select_state
        self.conflict_max_retries = max_retries
        self.conflict_retry_interval = retry_interval
        self.endpoint = endpoint

        super(SessionClient, self).__init__(**kwargs)

    def _parse_version_headers(self, resp):
        return self._generic_parse_version_headers(resp.headers.get)

    def _make_simple_request(self, conn, method, url):
        # NOTE: conn is self.session for this class
        return conn.request(url, method, raise_exc=False)

    @with_retries
    def _http_request(self, url, method, **kwargs):
        kwargs.setdefault('user_agent', USER_AGENT)
        kwargs.setdefault('auth', self.auth)
        if isinstance(self.endpoint_override, six.string_types):
            kwargs.setdefault(
                'endpoint_override',
                _trim_endpoint_api_version(self.endpoint_override)
            )

        if getattr(self, 'os_infra_optim_api_version', None):
            api_version = api_versioning.get_api_version(
                self.os_infra_optim_api_version)
            if api_version.is_latest():
                api_version = api_versioning.get_api_version(
                    LATEST_VERSION)
            kwargs['headers'].setdefault(
                'OpenStack-API-Version',
                ' '.join(['infra-optim',
                          api_version.get_string()]))

        endpoint_filter = kwargs.setdefault('endpoint_filter', {})
        endpoint_filter.setdefault('interface', self.interface)
        endpoint_filter.setdefault('service_type', self.service_type)
        endpoint_filter.setdefault('region_name', self.region_name)

        resp = self.session.request(url, method,
                                    raise_exc=False, **kwargs)
        if resp.status_code == http_client.NOT_ACCEPTABLE:
            negotiated_ver = self.negotiate_version(self.session, resp)
            kwargs['headers']['OpenStack-API-Version'] = (
                ' '.join(['infra-optim', negotiated_ver]))
            return self._http_request(url, method, **kwargs)
        if resp.status_code >= http_client.BAD_REQUEST:
            error_json = _extract_error_json(resp.content)
            raise exceptions.from_response(
                resp, error_json.get('faultstring'),
                error_json.get('debuginfo'), method, url)
        elif resp.status_code in (http_client.MOVED_PERMANENTLY,
                                  http_client.FOUND, http_client.USE_PROXY):
            # Redirected. Reissue the request to the new location.
            location = resp.headers.get('location')
            resp = self._http_request(location, method, **kwargs)
        elif resp.status_code == http_client.MULTIPLE_CHOICES:
            raise exceptions.from_response(resp, method=method, url=url)
        return resp

    def json_request(self, method, url, **kwargs):
        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Content-Type', 'application/json')
        kwargs['headers'].setdefault('Accept', 'application/json')

        if 'body' in kwargs:
            kwargs['data'] = jsonutils.dumps(kwargs.pop('body'))

        resp = self._http_request(url, method, **kwargs)
        body = resp.content
        content_type = resp.headers.get('content-type', None)
        status = resp.status_code
        if (status in (http_client.NO_CONTENT, http_client.RESET_CONTENT) or
                content_type is None):
            return resp, list()
        if 'application/json' in content_type:
            try:
                body = resp.json()
            except ValueError:
                LOG.error('Could not decode response body as JSON')
        else:
            body = None

        return resp, body

    def raw_request(self, method, url, **kwargs):
        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Content-Type',
                                     'application/octet-stream')
        return self._http_request(url, method, **kwargs)


def _construct_http_client(endpoint=None,
                           session=None,
                           token=None,
                           auth_ref=None,
                           os_infra_optim_api_version=DEFAULT_VER,
                           api_version_select_state='default',
                           max_retries=DEFAULT_MAX_RETRIES,
                           retry_interval=DEFAULT_RETRY_INTERVAL,
                           timeout=600,
                           ca_file=None,
                           cert_file=None,
                           key_file=None,
                           insecure=None,
                           **kwargs):
    if session:
        kwargs.setdefault('service_type', 'infra-optim')
        kwargs.setdefault('user_agent', 'python-watcherclient')
        kwargs.setdefault('interface', kwargs.pop('endpoint_type', None))
        kwargs.setdefault('endpoint_override', endpoint)

        ignored = {'token': token,
                   'auth_ref': auth_ref,
                   'timeout': timeout != 600,
                   'ca_file': ca_file,
                   'cert_file': cert_file,
                   'key_file': key_file,
                   'insecure': insecure}

        dvars = [k for k, v in ignored.items() if v]

        if dvars:
            LOG.warning('The following arguments are ignored when using '
                        'the session to construct a client: %s',
                        ', '.join(dvars))

        return SessionClient(
            session=session,
            os_infra_optim_api_version=os_infra_optim_api_version,
            api_version_select_state=api_version_select_state,
            max_retries=max_retries,
            retry_interval=retry_interval,
            endpoint=endpoint,
            **kwargs)
    else:
        if kwargs:
            LOG.warning('The following arguments are being ignored when '
                        'constructing the client: %s', ', '.join(kwargs))

        return HTTPClient(
            endpoint=endpoint,
            token=token,
            auth_ref=auth_ref,
            os_infra_optim_api_version=os_infra_optim_api_version,
            api_version_select_state=api_version_select_state,
            max_retries=max_retries,
            retry_interval=retry_interval,
            timeout=timeout,
            ca_file=ca_file,
            cert_file=cert_file,
            key_file=key_file,
            insecure=insecure)



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock

import testtools
from testtools import matchers

from oslo_utils.uuidutils import generate_uuid
from watcherclient.common.apiclient.exceptions import HTTPClientError
from watcherclient.tests.unit import utils
import watcherclient.v1.action_plan

ACTION_PLAN1 = {
    'id': 1,
    'uuid': 'f8e47706-efcf-49a4-a5c4-af604eb492f2',
    'audit': '770ef053-ecb3-48b0-85b5-d55a2dbc6588',
    'state': 'RECOMMENDED'
}

ACTION_PLAN2 = {
    'id': 2,
    'uuid': 'a5199d0e-0702-4613-9234-5ae2af8dafea',
    'audit': '239f02a5-9649-4e14-9d33-ac2bf67cb755',
    'state': 'RECOMMENDED'
}

UPDATED_ACTION_PLAN = copy.deepcopy(ACTION_PLAN1)
NEW_STATE = 'PENDING'
UPDATED_ACTION_PLAN['state'] = NEW_STATE

START_ACTION_PLAN = copy.deepcopy(ACTION_PLAN1)
START_ACTION_PLAN['state'] = NEW_STATE

ONGOING_ACTION_PLAN = copy.deepcopy(ACTION_PLAN1)
ONGOING_ACTION_PLAN['state'] = 'ONGOING'

CANCELLING_ACTION_PLAN = copy.deepcopy(ACTION_PLAN1)
CANCELLING_ACTION_PLAN['state'] = 'CANCELLING'

CANCELD_ACTION_PLAN = copy.deepcopy(ACTION_PLAN2)
CANCELD_ACTION_PLAN['state'] = 'CANCELLED'

fake_responses = {
    '/v1/action_plans':
    {
        'GET': (
            {},
            {"action_plans": [ACTION_PLAN1, ACTION_PLAN2]},
        ),
    },
    '/v1/action_plans/detail':
    {
        'GET': (
            {},
            {"action_plans": [ACTION_PLAN1, ACTION_PLAN2]},
        ),
    },
    '/v1/action_plans/%s' % ACTION_PLAN1['uuid']:
    {
        'GET': (
            {},
            ACTION_PLAN1,
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_ACTION_PLAN,
        ),
    },
    '/v1/action_plans/%s/start' % ACTION_PLAN1['uuid']:
    {
        'POST': (
            {},
            START_ACTION_PLAN,
        ),
    },
    '/v1/action_plans/detail?uuid=%s' % ACTION_PLAN1['uuid']:
    {
        'GET': (
            {},
            {"action_plans": [ACTION_PLAN1]},
        ),
    },
}

fake_responses_pagination = {
    '/v1/action_plans':
    {
        'GET': (
            {},
            {"action_plans": [ACTION_PLAN1],
             "next": "http://127.0.0.1:9322/v1/action_plans/?limit=1"}
        ),
    },
    '/v1/action_plans/?limit=1':
    {
        'GET': (
            {},
            {"action_plans": [ACTION_PLAN2]}
        ),
    },
    '/v1/action_plans/?marker=f8e47706-efcf-49a4-a5c4-af604eb492f2':
    {
        'GET': (
            {},
            {"action_plans": [ACTION_PLAN2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/action_plans/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"action_plans": [ACTION_PLAN2, ACTION_PLAN1]}
        ),
    },
    '/v1/action_plans/?sort_dir=desc':
    {
        'GET': (
            {},
            {"action_plans": [ACTION_PLAN2, ACTION_PLAN1]}
        ),
    },
}

fake_responses_cancel = {
    '/v1/action_plans/%s' % ACTION_PLAN1['uuid']:
    {
        'GET': (
            {},
            [ONGOING_ACTION_PLAN],
        ),
        'PATCH': (
            {},
            CANCELLING_ACTION_PLAN,
        ),
    },
    '/v1/action_plans/%s' % ACTION_PLAN2['uuid']:
    {
        'GET': (
            {},
            [ACTION_PLAN2],
        ),
        'PATCH': (
            {},
            CANCELD_ACTION_PLAN,
        ),
    },
}


class ActionPlanManagerTest(testtools.TestCase):

    def setUp(self):
        super(ActionPlanManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = watcherclient.v1.action_plan.ActionPlanManager(self.api)

    def test_action_plans_list(self):
        action_plans = self.mgr.list()
        expect = [
            ('GET', '/v1/action_plans', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(action_plans))

    def test_action_plans_list_detail(self):
        action_plans = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/action_plans/detail', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(action_plans))

    def test_action_plans_list_limit(self):
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = watcherclient.v1.action_plan.ActionPlanManager(self.api)
        action_plans = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/action_plans/?limit=1', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(action_plans, matchers.HasLength(1))

    def test_action_plans_list_pagination_no_limit(self):
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = watcherclient.v1.action_plan.ActionPlanManager(self.api)
        action_plans = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/action_plans', {}, None),
            ('GET', '/v1/action_plans/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(action_plans, matchers.HasLength(2))

    def test_action_plans_list_marker(self):
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = watcherclient.v1.action_plan.ActionPlanManager(self.api)
        action_plans = self.mgr.list(
            marker='f8e47706-efcf-49a4-a5c4-af604eb492f2')
        expect = [
            ('GET', '/v1/action_plans/?'
             'marker=f8e47706-efcf-49a4-a5c4-af604eb492f2',
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(action_plans, matchers.HasLength(1))

    def test_action_plans_list_sort_key(self):
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = watcherclient.v1.action_plan.ActionPlanManager(self.api)
        action_plans = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/action_plans/?sort_key=updated_at', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(action_plans))

    def test_action_plans_list_sort_dir(self):
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = watcherclient.v1.action_plan.ActionPlanManager(self.api)
        action_plans = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/action_plans/?sort_dir=desc', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(action_plans))

    def test_action_plans_show(self):
        action_plan = self.mgr.get(ACTION_PLAN1['uuid'])
        expect = [
            ('GET', '/v1/action_plans/%s' % ACTION_PLAN1['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(ACTION_PLAN1['uuid'], action_plan.uuid)

    def test_action_plans_get_index_error(self):

        # verify this method will return None when meet indexError
        fake_uuid = generate_uuid()
        self.api.json_request = mock.Mock(return_value=('404', []))
        self.assertIsNone(self.mgr.get(fake_uuid))

    def test_action_plans_delete(self):
        # verity that action plan was successfully deleted
        self.api.raw_request = mock.Mock(return_value=('204', []))
        self.assertIsNone(self.mgr.delete(ACTION_PLAN1['uuid']))

        # verity that delete a wrong action plan will raise Exception
        fake_uuid = generate_uuid()
        self.api.raw_request = mock.Mock(
            side_effect=HTTPClientError('404 Not Found'))
        self.assertRaises(HTTPClientError, self.mgr.delete, fake_uuid)

    def test_action_plans_cancel(self):
        # verity that the status of action plan can be converted from
        # 'ONGOING' to 'CANCELLING'
        self.api = utils.FakeAPI(fake_responses_cancel)
        self.mgr = watcherclient.v1.action_plan.ActionPlanManager(self.api)
        patch = {'op': 'replace',
                 'value': 'CANCELLING',
                 'path': '/state'}
        action_plan = self.mgr.cancel(action_plan_id=ACTION_PLAN1['uuid'])
        expect = [
            ('GET', '/v1/action_plans/%s' % ACTION_PLAN1['uuid'], {}, None),
            ('PATCH', '/v1/action_plans/%s' % ACTION_PLAN1['uuid'], {},
             [patch])
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual('CANCELLING', action_plan.state)

        # verity that the status of action plan can be converted from
        # 'RECOMMENDED' to 'CANCELLED'
        patch['value'] = 'CANCELLED'
        self.api.calls = []
        action_plan = self.mgr.cancel(action_plan_id=ACTION_PLAN2['uuid'])
        expect = [
            ('GET', '/v1/action_plans/%s' % ACTION_PLAN2['uuid'], {}, None),
            ('PATCH', '/v1/action_plans/%s' % ACTION_PLAN2['uuid'], {},
             [patch])
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual('CANCELLED', action_plan.state)

    def test_action_plan_update(self):
        patch = {'op': 'replace',
                 'value': NEW_STATE,
                 'path': '/state'}
        action_plan = self.mgr.update(action_plan_id=ACTION_PLAN1['uuid'],
                                      patch=patch)
        expect = [
            ('PATCH',
             '/v1/action_plans/%s' % ACTION_PLAN1['uuid'],
             {},
             patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(NEW_STATE, action_plan.state)

    def test_action_plan_start(self):
        action_plan = self.mgr.start(ACTION_PLAN1['uuid'])
        expect = [('POST', '/v1/action_plans/%s/start'
                   % ACTION_PLAN1['uuid'], {}, {})]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(NEW_STATE, action_plan.state)



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from watcherclient._i18n import _
from watcherclient.common import httpclient
from watcherclient import exceptions
from watcherclient import v1


class Client(object):
    """Client for the Watcher v1 API.

    :param string endpoint: A user-supplied endpoint URL for the watcher
                            service.
    :param function token: Provides token for authentication.
    :param integer timeout: Allows customization of the timeout for client
                            http requests. (optional)
    """

    def __init__(self, endpoint=None, *args, **kwargs):
        """Initialize a new client for the Watcher v1 API."""
        if kwargs.get('os_infra_optim_api_version'):
            kwargs['api_version_select_state'] = "user"
        else:
            if not endpoint:
                raise exceptions.EndpointException(
                    _("Must provide 'endpoint' if os_infra_optim_api_version "
                      "isn't specified"))

            # If the user didn't specify a version, use a cached version if
            # one has been stored
            host, netport = httpclient.get_server(endpoint)
            kwargs['api_version_select_state'] = "default"
            kwargs['os_infra_optim_api_version'] = httpclient.DEFAULT_VER

        self.http_client = httpclient._construct_http_client(
            endpoint, *args, **kwargs)

        self.audit = v1.AuditManager(self.http_client)
        self.audit_template = v1.AuditTemplateManager(self.http_client)
        self.action = v1.ActionManager(self.http_client)
        self.action_plan = v1.ActionPlanManager(self.http_client)
        self.goal = v1.GoalManager(self.http_client)
        self.scoring_engine = v1.ScoringEngineManager(self.http_client)
        self.service = v1.ServiceManager(self.http_client)
        self.strategy = v1.StrategyManager(self.http_client)
        self.data_model = v1.DataModelManager(self.http_client)



#
# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from watcherclient.common import base
from watcherclient.common import utils
# from watcherclient import exceptions as exc


class ActionPlan(base.Resource):
    def __repr__(self):
        return "<ActionPlan %s>" % self._info


class ActionPlanManager(base.Manager):
    resource_class = ActionPlan

    @staticmethod
    def _path(id=None, q_param=None):
        if id and q_param:
            return '/v1/action_plans/%s/%s' % (id, q_param)
        elif id:
            return '/v1/action_plans/%s' % id
        else:
            return '/v1/action_plans'

    def list(self, audit=None, limit=None, sort_key=None,
             sort_dir=None, detail=False, marker=None):
        """Retrieve a list of action plan.

        :param audit: Name of the audit
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of action plans to return.
            2) limit == 0, return the entire list of action plans.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Watcher API
               (see Watcher's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about action plans.

        :param marker: The last actionplan UUID of the previous page;
                       displays list of actionplans after "marker".

        :returns: A list of action plans.

        """
        if limit is not None:
            limit = int(limit)

        filters = utils.common_filters(limit, sort_key, sort_dir, marker)
        if audit is not None:
            filters.append('audit_uuid=%s' % audit)

        path = ''
        if detail:
            path += 'detail'
        if filters:
            path += '?' + '&'.join(filters)

        if limit is None:
            return self._list(self._path(path), "action_plans")
        else:
            return self._list_pagination(self._path(path), "action_plans",
                                         limit=limit)

    def get(self, action_plan_id):
        try:
            return self._list(self._path(action_plan_id))[0]
        except IndexError:
            return None

    def delete(self, action_plan_id):
        return self._delete(self._path(action_plan_id))

    def update(self, action_plan_id, patch):
        return self._update(self._path(action_plan_id), patch)

    def start(self, action_plan_id):
        return self._start(self._path(action_plan_id, 'start'))

    def cancel(self, action_plan_id):
        action_plan = self.get(action_plan_id)
        if action_plan.state == "ONGOING":
            patch = [{'op': 'replace', 'value': 'CANCELLING',
                      'path': '/state'}]
        else:
            patch = [{'op': 'replace', 'value': 'CANCELLED', 'path': '/state'}]
        return self._update(self._path(action_plan_id), patch)



#   Copyright 2016 NEC Corporation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import abc
import logging

from cliff import command
from cliff import lister
from cliff import show
import six


class CommandMeta(abc.ABCMeta):

    def __new__(mcs, name, bases, cls_dict):
        if 'log' not in cls_dict:
            cls_dict['log'] = logging.getLogger(
                cls_dict['__module__'] + '.' + name)
        return super(CommandMeta, mcs).__new__(mcs, name, bases, cls_dict)


@six.add_metaclass(CommandMeta)
class Command(command.Command):

    def run(self, parsed_args):
        self.log.debug('run(%s)', parsed_args)
        return super(Command, self).run(parsed_args)


class Lister(Command, lister.Lister):
    pass


class ShowOne(Command, show.ShowOne):
    def get_parser(self, prog_name, formatter_class=None):
        parser = super(ShowOne, self).get_parser(prog_name)
        if formatter_class:
            parser.formatter_class = formatter_class
        return parser



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as kaloading

from oslo_utils import importutils

from watcherclient._i18n import _
from watcherclient.common import api_versioning
from watcherclient import exceptions


def get_client(api_version, os_auth_token=None, watcher_url=None,
               os_username=None, os_password=None, os_auth_url=None,
               os_project_id=None, os_project_name=None, os_tenant_id=None,
               os_tenant_name=None, os_region_name=None,
               os_user_domain_id=None, os_user_domain_name=None,
               os_project_domain_id=None, os_project_domain_name=None,
               os_service_type=None, os_endpoint_type=None,
               insecure=None, timeout=None, os_cacert=None, ca_file=None,
               os_cert=None, cert_file=None, os_key=None, key_file=None,
               os_infra_optim_api_version=None, max_retries=None,
               retry_interval=None, session=None, os_endpoint_override=None,
               **ignored_kwargs):
    """Get an authenticated client, based on the credentials.

    :param api_version: the API version to use. Valid value: '1'.
    :param os_auth_token: pre-existing token to re-use
    :param watcher_url: watcher API endpoint
    :param os_username: name of a user
    :param os_password: user's password
    :param os_auth_url: endpoint to authenticate against
    :param os_project_id: ID of a project
    :param os_project_name: name of a project
    :param os_tenant_id: ID of a tenant (deprecated in favour of
        os_project_id)
    :param os_tenant_name: name of a tenant (deprecated in favour of
        os_project_name)
    :param os_region_name: name of a keystone region
    :param os_user_domain_id: ID of a domain the user belongs to
    :param os_user_domain_name: name of a domain the user belongs to
    :param os_project_domain_id: ID of a domain the project belongs to
    :param os_project_domain_name: name of a domain the project belongs to
    :param os_service_type: the type of service to lookup the endpoint for
    :param os_endpoint_type: the type (exposure) of the endpoint
    :param insecure: allow insecure SSL (no cert verification)
    :param timeout: allows customization of the timeout for client HTTP
        requests
    :param os_cacert: path to cacert file
    :param ca_file: path to cacert file, deprecated in favour of os_cacert
    :param os_cert: path to cert file
    :param cert_file: path to cert file, deprecated in favour of os_cert
    :param os_key: path to key file
    :param key_file: path to key file, deprecated in favour of os_key
    :param os_infra_optim_api_version: watcher API version to use
    :param max_retries: Maximum number of retries in case of conflict error
    :param retry_interval: Amount of time (in seconds) between retries in case
        of conflict error
    :param session: Keystone session to use
    :param os_endpoint_override: watcher API endpoint
    :param ignored_kwargs: all the other params that are passed. Left for
        backwards compatibility. They are ignored.
    """
    os_service_type = os_service_type or 'infra-optim'
    os_endpoint_type = os_endpoint_type or 'publicURL'
    project_id = (os_project_id or os_tenant_id)
    project_name = (os_project_name or os_tenant_name)
    kwargs = {
        'os_infra_optim_api_version': os_infra_optim_api_version,
        'max_retries': max_retries,
        'retry_interval': retry_interval,
    }
    endpoint = watcher_url or os_endpoint_override
    cacert = os_cacert or ca_file
    cert = os_cert or cert_file
    key = os_key or key_file
    if os_auth_token and endpoint:
        kwargs.update({
            'token': os_auth_token,
            'insecure': insecure,
            'ca_file': cacert,
            'cert_file': cert,
            'key_file': key,
            'timeout': timeout,
        })
    elif os_auth_url:
        auth_type = 'password'
        auth_kwargs = {
            'auth_url': os_auth_url,
            'project_id': project_id,
            'project_name': project_name,
            'user_domain_id': os_user_domain_id,
            'user_domain_name': os_user_domain_name,
            'project_domain_id': os_project_domain_id,
            'project_domain_name': os_project_domain_name,
        }
        if os_username and os_password:
            auth_kwargs.update({
                'username': os_username,
                'password': os_password,
            })
        elif os_auth_token:
            auth_type = 'token'
            auth_kwargs.update({
                'token': os_auth_token,
            })
        # Create new session only if it was not passed in
        if not session:
            loader = kaloading.get_plugin_loader(auth_type)
            auth_plugin = loader.load_from_options(**auth_kwargs)
            # Let keystoneauth do the necessary parameter conversions
            session = kaloading.session.Session().load_from_options(
                auth=auth_plugin, insecure=insecure, cacert=cacert,
                cert=cert, key=key, timeout=timeout,
            )

    exception_msg = _('Must provide Keystone credentials or user-defined '
                      'endpoint and token')
    if not endpoint:
        if session:
            try:
                # Pass the endpoint, it will be used to get hostname
                # and port that will be used for API version caching. It will
                # be also set as endpoint_override.
                endpoint = session.get_endpoint(
                    service_type=os_service_type,
                    interface=os_endpoint_type,
                    region_name=os_region_name
                )
            except Exception as e:
                raise exceptions.AmbiguousAuthSystem(
                    exception_msg + _(', error was: %s') % e)
        else:
            # Neither session, nor valid auth parameters provided
            raise exceptions.AmbiguousAuthSystem(exception_msg)

    # Always pass the session
    kwargs['session'] = session

    return Client(api_version, endpoint, **kwargs)


def _get_client_class_and_version(version):
    if not isinstance(version, api_versioning.APIVersion):
        version = api_versioning.get_api_version(version)
    else:
        api_versioning.check_major_version(version)
    if version.is_latest():
        raise exceptions.UnsupportedVersion(
            _("The version should be explicit, not latest."))
    return version, importutils.import_class(
        "watcherclient.v%s.client.Client" % version.ver_major)


def Client(version, *args, **kwargs):
    """Initialize client object based on given version.

    HOW-TO:
    The simplest way to create a client instance is initialization with your
    credentials::

        >>> from watcherclient import client
        >>> watcher = client.Client(VERSION, USERNAME, PASSWORD,
        ...                      PROJECT_ID, AUTH_URL)

    Here ``VERSION`` can be a string or
    ``watcherclient.api_versions.APIVersion`` obj. If you prefer string value,
    you can use ``1`` or ``1.X`` (where X is a microversion).


    Alternatively, you can create a client instance using the keystoneauth
    session API. See "The watcherclient Python API" page at
    python-watcherclient's doc.
    """
    api_version, client_class = _get_client_class_and_version(version)

    kw_api = kwargs.get('os_infra_optim_api_version')
    endpoint = kwargs.get('endpoint')
    # If both os_infra_optim_api_version and endpoint are not provided, get
    # API version from arg.
    if not kw_api and not endpoint:
        kwargs['os_infra_optim_api_version'] = api_version.get_string()
    return client_class(*args, **kwargs)



# Copyright (c) 2016 Intel
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from osc_lib import utils

from watcherclient._i18n import _
from watcherclient.common import command
from watcherclient.common import utils as common_utils
from watcherclient import exceptions
from watcherclient.v1 import resource_fields as res_fields


class ShowScoringEngine(command.ShowOne):
    """Show detailed information about a given scoring engine."""

    def get_parser(self, prog_name):
        parser = super(ShowScoringEngine, self).get_parser(prog_name)
        parser.add_argument(
            'scoring_engine',
            metavar='<scoring_engine>',
            help=_('Name of the scoring engine'),
        )
        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        try:
            scoring_engine = client.scoring_engine.get(
                parsed_args.scoring_engine)
        except exceptions.HTTPNotFound as exc:
            raise exceptions.CommandError(str(exc))

        columns = res_fields.SCORING_ENGINE_FIELDS
        column_headers = res_fields.SCORING_ENGINE_FIELD_LABELS

        return column_headers, utils.get_item_properties(scoring_engine,
                                                         columns)


class ListScoringEngine(command.Lister):
    """List information on retrieved scoring engines."""

    def get_parser(self, prog_name):
        parser = super(ListScoringEngine, self).get_parser(prog_name)
        parser.add_argument(
            '--detail',
            dest='detail',
            action='store_true',
            default=False,
            help=_("Show detailed information about scoring engines."))
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of actions to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Watcher API Service.'))
        parser.add_argument(
            '--sort-key',
            metavar='<field>',
            help=_('Action field that will be used for sorting.'))
        parser.add_argument(
            '--sort-dir',
            metavar='<direction>',
            choices=['asc', 'desc'],
            help=_('Sort direction: "asc" (the default) or "desc".'))

        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        params = {}
        if parsed_args.detail:
            fields = res_fields.SCORING_ENGINE_FIELDS
            field_labels = res_fields.SCORING_ENGINE_FIELD_LABELS
        else:
            fields = res_fields.SCORING_ENGINE_SHORT_LIST_FIELDS
            field_labels = res_fields.SCORING_ENGINE_SHORT_LIST_FIELD_LABELS

        params.update(
            common_utils.common_params_for_list(
                parsed_args, fields, field_labels))

        try:
            data = client.scoring_engine.list(**params)
        except exceptions.HTTPNotFound as ex:
            raise exceptions.CommandError(str(ex))

        return (field_labels,
                (utils.get_item_properties(item, fields) for item in data))



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import os
from unittest import mock

import fixtures
from oslo_utils import strutils
import six
import testtools


class BaseTestCase(testtools.TestCase):

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.useFixture(fixtures.FakeLogger())

        # If enabled, stdout and/or stderr is captured and will appear in
        # test results if that test fails.
        if strutils.bool_from_string(os.environ.get('OS_STDOUT_CAPTURE')):
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if strutils.bool_from_string(os.environ.get('OS_STDERR_CAPTURE')):
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))


class FakeAPI(object):
    def __init__(self, responses):
        self.responses = responses
        self.calls = []

    def _request(self, method, url, headers=None, body=None):
        call = (method, url, headers or {}, body)
        self.calls.append(call)
        return self.responses[url][method]

    def raw_request(self, *args, **kwargs):
        response = self._request(*args, **kwargs)
        body_iter = iter(six.StringIO(response[1]))
        return FakeResponse(response[0]), body_iter

    def json_request(self, *args, **kwargs):
        response = self._request(*args, **kwargs)
        return FakeResponse(response[0]), response[1]


class FakeConnection(object):
    def __init__(self, response=None):
        self._response = response
        self._last_request = None

    def request(self, method, conn_url, **kwargs):
        self._last_request = (method, conn_url, kwargs)

    def setresponse(self, response):
        self._response = response

    def getresponse(self):
        return self._response

    def __repr__(self):
        return ("FakeConnection(response=%s)" % (self._response))


class FakeResponse(object):
    def __init__(self, headers, body=None, version=None, status=None,
                 reason=None):
        """Fake object to help testing.

        :param headers: dict representing HTTP response headers
        :param body: file-like object
        """
        self.headers = headers
        self.body = body
        self.raw = mock.Mock()
        self.raw.version = version
        self.status_code = status
        self.reason = reason

    def getheaders(self):
        return copy.deepcopy(self.headers).items()

    def getheader(self, key, default):
        return self.headers.get(key, default)

    def read(self, amt):
        return self.body.read(amt)

    def __repr__(self):
        return ("FakeResponse(%s, body=%s, version=%s, status=%s, reason=%s)" %
                (self.headers, self.body, self.version, self.status,
                 self.reason))


class FakeSessionResponse(object):

    def __init__(self, headers, content=None, status_code=None, version=None):
        self.headers = headers
        self.content = content
        self.status_code = status_code
        self.raw = mock.Mock()
        self.raw.version = version
        self.reason = ''

    def iter_content(self, chunk_size):
        return iter(self.content)


class FakeSession(object):

    def __init__(self, headers, content=None, status_code=None, version=None):
        self.headers = headers
        self.content = content
        self.status_code = status_code
        self.version = version
        self.verify = False
        self.cert = ('test_cert', 'test_key')

    def request(self, url, method, **kwargs):
        request = FakeSessionResponse(
            self.headers, self.content, self.status_code, self.version)
        return request



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import testtools
from testtools import matchers

from watcherclient.tests.unit import utils
import watcherclient.v1.strategy

STRATEGY1 = {
    'uuid': '2cf86250-d309-4b81-818e-1537f3dba6e5',
    'name': 'basic',
    'display_name': 'Basic consolidation',
    'strategy_id': 'SERVER_CONSOLIDATION',
}

STRATEGY2 = {
    'uuid': 'b20bb987-ea8f-457a-a4ea-ab3ffdfeff8b',
    'name': 'dummy',
    'display_name': 'Dummy',
    'strategy_id': 'DUMMY',
}

fake_responses = {
    '/v1/strategies':
    {
        'GET': (
            {},
            {"strategies": [STRATEGY1]},
        ),
    },
    '/v1/strategies/detail':
    {
        'GET': (
            {},
            {"strategies": [STRATEGY1]},
        )
    },
    '/v1/strategies/%s' % STRATEGY1['uuid']:
    {
        'GET': (
            {},
            STRATEGY1,
        ),
    },
    '/v1/strategies/%s' % STRATEGY1['name']:
    {
        'GET': (
            {},
            STRATEGY1,
        ),
    },
    '/v1/strategies/%s/state' % STRATEGY1['name']:
    {
        'GET': (
            {},
            STRATEGY1,
        ),
    },
}

fake_responses_pagination = {
    '/v1/strategies':
    {
        'GET': (
            {},
            {"strategies": [STRATEGY1],
             "next": "http://127.0.0.1:9322/v1/strategies/?limit=1"}
        ),
    },
    '/v1/strategies/?limit=1':
    {
        'GET': (
            {},
            {"strategies": [STRATEGY2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/strategies/?sort_key=id':
    {
        'GET': (
            {},
            {"strategies": [STRATEGY1, STRATEGY2]}
        ),
    },
    '/v1/strategies/?sort_dir=desc':
    {
        'GET': (
            {},
            {"strategies": [STRATEGY2, STRATEGY1]}
        ),
    },
}

fake_responses_marker = {
    '/v1/strategies/?marker=2cf86250-d309-4b81-818e-1537f3dba6e5':
    {
        'GET': (
            {},
            {"strategies": [STRATEGY2]}
        ),
    },
}


class StrategyManagerTest(testtools.TestCase):

    def setUp(self):
        super(StrategyManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = watcherclient.v1.strategy.StrategyManager(self.api)

    def test_strategies_list(self):
        strategies = self.mgr.list()
        expect = [
            ('GET', '/v1/strategies', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(strategies))

    def test_strategies_list_detail(self):
        strategies = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/strategies/detail', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(strategies))

    def test_strategies_list_marker(self):
        self.api = utils.FakeAPI(fake_responses_marker)
        self.mgr = watcherclient.v1.strategy.StrategyManager(self.api)
        strategies = self.mgr.list(marker=STRATEGY1['uuid'])
        expect = [
            ('GET',
             '/v1/strategies/?marker=2cf86250-d309-4b81-818e-1537f3dba6e5',
             {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(strategies))

    def test_strategies_list_limit(self):
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = watcherclient.v1.strategy.StrategyManager(self.api)
        strategies = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/strategies/?limit=1', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(strategies, matchers.HasLength(1))

    def test_strategies_list_pagination_no_limit(self):
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = watcherclient.v1.strategy.StrategyManager(self.api)
        strategies = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/strategies', {}, None),
            ('GET', '/v1/strategies/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(strategies, matchers.HasLength(2))

    def test_strategies_list_sort_key(self):
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = watcherclient.v1.strategy.StrategyManager(self.api)
        strategies = self.mgr.list(sort_key='id')
        expect = [
            ('GET', '/v1/strategies/?sort_key=id', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(strategies))

    def test_strategies_list_sort_dir(self):
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = watcherclient.v1.strategy.StrategyManager(self.api)
        strategies = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/strategies/?sort_dir=desc', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(strategies))

    def test_strategies_show(self):
        strategy = self.mgr.get(STRATEGY1['uuid'])
        expect = [
            ('GET', '/v1/strategies/%s' % STRATEGY1['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(STRATEGY1['uuid'], strategy.uuid)

    def test_strategies_show_by_name(self):
        strategy = self.mgr.get(STRATEGY1['name'])
        expect = [
            ('GET', '/v1/strategies/%s' % STRATEGY1['name'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(STRATEGY1['name'], strategy.name)

    def test_strategies_state(self):
        self.mgr.state(STRATEGY1['name'])
        expect = [
            ('GET', '/v1/strategies/%s/state' % STRATEGY1['name'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shlex
from unittest import mock

from osc_lib import utils as oscutils
from oslo_serialization import jsonutils

from watcherclient.common import httpclient
from watcherclient.tests.unit import utils


class CommandTestCase(utils.BaseTestCase):

    def setUp(self, os_infra_optim_api_version='1.0'):
        super(CommandTestCase, self).setUp()

        self.fake_env = {
            'debug': False,
            'insecure': False,
            'no_auth': False,
            'os_auth_token': '',
            'os_auth_url': 'http://127.0.0.1:5000/v2.0',
            'os_endpoint_override': 'http://watcher-endpoint:9322',
            'os_username': 'test',
            'os_password': 'test',
            'timeout': 600,
            'os_infra_optim_api_version': os_infra_optim_api_version}
        self.m_env = mock.Mock(
            name='m_env',
            side_effect=lambda x, *args, **kwargs: self.fake_env.get(
                x.lower(), kwargs.get('default', '')))
        self.p_env = mock.patch.object(oscutils, 'env', self.m_env)
        self.p_env.start()
        self.addCleanup(self.p_env.stop)

        self.p_construct_http_client = mock.patch.object(
            httpclient, '_construct_http_client')
        self.m_construct_http_client = self.p_construct_http_client.start()
        self.addCleanup(self.p_construct_http_client.stop)

    def run_cmd(self, cmd, formatting='json'):
        if formatting and formatting != 'table':
            formatter_arg = " -f %s" % formatting
            formatter = jsonutils.loads
        else:
            formatter_arg = ''
            formatter = str
        formatted_cmd = "%(cmd)s%(formatter)s" % dict(
            cmd=cmd, formatter=formatter_arg)

        exit_code = self.cmd.run(shlex.split(formatted_cmd))

        try:
            raw_data = self.stdout.getvalue()
            formatted_output = formatter(self.stdout.getvalue())
        except Exception:
            self.fail("Formatting error (`%s` -> '%s')" %
                      (raw_data, formatting))
        return exit_code, formatted_output

    def resource_as_dict(self, resource, columns=(), column_headers=()):
        mapping = dict(zip(columns, column_headers))
        return {mapping[k]: v for k, v in resource.to_dict().items()
                if not columns or columns and k in mapping}



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from six.moves.urllib import parse as urlparse
from testtools import matchers

from watcherclient.tests.unit import utils
import watcherclient.v1.audit_template

AUDIT_TMPL1 = {
    'id': 1,
    'uuid': 'f8e47706-efcf-49a4-a5c4-af604eb492f2',
    'name': 'Audit Template 1',
    'description': 'Audit Template 1 description',
    'goal_uuid': '7568667b-51fe-4087-9eb1-29b26891036f',
    'goal_name': 'SERVER_CONSOLIDATION',
    'strategy_uuid': 'bbe6b966-f98e-439b-a01a-17b9b3b8478b',
    'strategy_name': 'server_consolidation',
}

AUDIT_TMPL2 = {
    'id': 2,
    'uuid': 'a5199d0e-0702-4613-9234-5ae2af8dafea',
    'name': 'Audit Template 2',
    'description': 'Audit Template 2 description',
    'goal_uuid': 'e75ee410-b32b-465f-88b5-4397705f9473',
    'goal_name': 'DUMMY',
    'strategy_uuid': 'ae99a4a4-acbc-4c67-abe1-e37128fac45d',
    'strategy_name': 'dummy',
}

AUDIT_TMPL3 = {
    'id': 3,
    'uuid': '770ef053-ecb3-48b0-85b5-d55a2dbc6588',
    'name': 'Audit Template 3',
    'description': 'Audit Template 3 description',
    'goal_uuid': '7568667b-51fe-4087-9eb1-29b26891036f',
    'goal_name': 'SERVER_CONSOLIDATION',
}

CREATE_AUDIT_TEMPLATE = copy.deepcopy(AUDIT_TMPL1)
del CREATE_AUDIT_TEMPLATE['id']
del CREATE_AUDIT_TEMPLATE['uuid']
del CREATE_AUDIT_TEMPLATE['goal_name']
del CREATE_AUDIT_TEMPLATE['strategy_name']
CREATE_AUDIT_TEMPLATE['goal'] = CREATE_AUDIT_TEMPLATE.pop('goal_uuid')
CREATE_AUDIT_TEMPLATE['strategy'] = CREATE_AUDIT_TEMPLATE.pop('strategy_uuid')

UPDATED_AUDIT_TMPL1 = copy.deepcopy(AUDIT_TMPL1)
NEW_NAME = 'Audit Template_1 new name'
UPDATED_AUDIT_TMPL1['name'] = NEW_NAME

fake_responses = {
    '/v1/audit_templates':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL1]},
        ),
        'POST': (
            {},
            CREATE_AUDIT_TEMPLATE,
        ),
    },
    '/v1/audit_templates/detail':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL1]},
        ),
    },
    '/v1/audit_templates/%s' % AUDIT_TMPL1['uuid']:
    {
        'GET': (
            {},
            AUDIT_TMPL1,
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_AUDIT_TMPL1,
        ),
    },
    urlparse.quote('/v1/audit_templates/%s' % AUDIT_TMPL1['name']):
    {
        'GET': (
            {},
            AUDIT_TMPL1,
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_AUDIT_TMPL1,
        ),
    },
    '/v1/audit_templates/detail?name=%s' % AUDIT_TMPL1['name']:
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL1]},
        ),
    },
    '/v1/audit_templates/?name=%s' % AUDIT_TMPL1['name']:
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL1]},
        ),
    },
    '/v1/audit_templates/detail?goal=%s' % AUDIT_TMPL1['goal_uuid']:
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL1, AUDIT_TMPL3]},
        ),
    },
    '/v1/audit_templates/?goal=%s' % AUDIT_TMPL1['goal_uuid']:
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL1, AUDIT_TMPL3]},
        ),
    }
}

fake_responses_pagination = {
    '/v1/audit_templates':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL1],
             "next": "http://127.0.0.1:9322/v1/audit_templates/?limit=1"}
        ),
    },
    '/v1/audit_templates/?limit=1':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/audit_templates/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL3, AUDIT_TMPL2, AUDIT_TMPL1]}
        ),
    },
    '/v1/audit_templates/?sort_dir=desc':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL3, AUDIT_TMPL2, AUDIT_TMPL1]}
        ),
    },
}

fake_responses_marker = {
    '/v1/audit_templates/?marker=f8e47706-efcf-49a4-a5c4-af604eb492f2':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL2, AUDIT_TMPL3]}
        ),
    },
}

fake_responses_filter_by_goal_uuid = {
    '/v1/audit_templates/?goal=e75ee410-b32b-465f-88b5-4397705f9473':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL2]}
        ),
    },
}

fake_responses_filter_by_goal_name = {
    '/v1/audit_templates/?goal=DUMMY':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL2]}
        ),
    },
}

fake_responses_filter_by_strategy_uuid = {
    '/v1/audit_templates/?strategy=ae99a4a4-acbc-4c67-abe1-e37128fac45d':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL2]}
        ),
    },
}

fake_responses_filter_by_strategy_name = {
    '/v1/audit_templates/?strategy=dummy':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL2]}
        ),
    },
}

fake_responses_filter_by_strategy_and_goal_name = {
    '/v1/audit_templates/?goal=DUMMY&strategy=dummy':
    {
        'GET': (
            {},
            {"audit_templates": [AUDIT_TMPL2]}
        ),
    },
}


class AuditTemplateManagerTest(utils.BaseTestCase):

    def setUp(self):
        super(AuditTemplateManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)

    def test_audit_templates_list(self):
        audit_templates = self.mgr.list()
        expect = [
            ('GET', '/v1/audit_templates', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_filter_by_name(self):
        audit_templates = self.mgr.list(name=AUDIT_TMPL1['name'])
        expect = [
            ('GET', '/v1/audit_templates/?name=%s' % AUDIT_TMPL1['name'],
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_filter_by_goal_uuid(self):
        self.api = utils.FakeAPI(fake_responses_filter_by_goal_uuid)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(
            goal="e75ee410-b32b-465f-88b5-4397705f9473")
        expect = [
            ('GET',
             '/v1/audit_templates/?goal=%s' % AUDIT_TMPL2['goal_uuid'],
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_filter_by_goal_name(self):
        self.api = utils.FakeAPI(fake_responses_filter_by_goal_name)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(goal="DUMMY")
        expect = [
            ('GET',
             '/v1/audit_templates/?goal=%s' % AUDIT_TMPL2['goal_name'],
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_filter_by_strategy_uuid(self):
        self.api = utils.FakeAPI(fake_responses_filter_by_strategy_uuid)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(
            strategy="ae99a4a4-acbc-4c67-abe1-e37128fac45d")
        expect = [
            ('GET',
             '/v1/audit_templates/?strategy=%s' % (
                 AUDIT_TMPL2['strategy_uuid']),
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_filter_by_strategy_name(self):
        self.api = utils.FakeAPI(fake_responses_filter_by_strategy_name)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(strategy="dummy")
        expect = [
            ('GET',
             '/v1/audit_templates/?strategy=%s' % (
                 AUDIT_TMPL2['strategy_name']),
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_filter_by_goal_and_strategy_name(self):
        self.api = utils.FakeAPI(
            fake_responses_filter_by_strategy_and_goal_name)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(goal="DUMMY", strategy="dummy")
        expect = [
            ('GET',
             '/v1/audit_templates/?goal=%s&strategy=%s' % (
                 AUDIT_TMPL2['goal_name'], AUDIT_TMPL2['strategy_name']),
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_detail(self):
        audit_templates = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/audit_templates/detail', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_by_name_detail(self):
        audit_templates = self.mgr.list(name=AUDIT_TMPL1['name'], detail=True)
        expect = [
            ('GET',
             '/v1/audit_templates/detail?name=%s' % AUDIT_TMPL1['name'],
             {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(audit_templates))

    def test_audit_templates_list_limit(self):
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/audit_templates/?limit=1', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(audit_templates, matchers.HasLength(1))

    def test_audit_templates_list_pagination_no_limit(self):
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/audit_templates', {}, None),
            ('GET', '/v1/audit_templates/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(audit_templates, matchers.HasLength(2))

    def test_audit_templates_list_sort_key(self):
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/audit_templates/?sort_key=updated_at', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(3, len(audit_templates))

    def test_audit_templates_list_sort_dir(self):
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/audit_templates/?sort_dir=desc', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(3, len(audit_templates))

    def test_audit_templates_list_marker(self):
        self.api = utils.FakeAPI(fake_responses_marker)
        self.mgr = watcherclient.v1.audit_template.AuditTemplateManager(
            self.api)
        audit_templates = self.mgr.list(marker=AUDIT_TMPL1['uuid'])
        expect_url = '/v1/audit_templates/?marker=%s' % AUDIT_TMPL1['uuid']
        expect = [
            ('GET', expect_url, {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(audit_templates))

    def test_audit_templates_show(self):
        audit_template = self.mgr.get(AUDIT_TMPL1['uuid'])
        expect = [
            ('GET', '/v1/audit_templates/%s' % AUDIT_TMPL1['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(AUDIT_TMPL1['uuid'], audit_template.uuid)
        self.assertEqual(AUDIT_TMPL1['name'], audit_template.name)
        self.assertEqual(AUDIT_TMPL1['description'],
                         audit_template.description)
        self.assertEqual(AUDIT_TMPL1['goal_uuid'], audit_template.goal_uuid)
        self.assertEqual(AUDIT_TMPL1['strategy_uuid'],
                         audit_template.strategy_uuid)

    def test_audit_templates_show_by_name(self):
        audit_template = self.mgr.get(urlparse.quote(AUDIT_TMPL1['name']))
        expect = [
            ('GET',
             urlparse.quote('/v1/audit_templates/%s' % AUDIT_TMPL1['name']),
             {}, None),
        ]

        self.assertEqual(expect, self.api.calls)
        self.assertEqual(AUDIT_TMPL1['uuid'], audit_template.uuid)
        self.assertEqual(AUDIT_TMPL1['name'], audit_template.name)
        self.assertEqual(AUDIT_TMPL1['description'],
                         audit_template.description)
        self.assertEqual(AUDIT_TMPL1['goal_uuid'], audit_template.goal_uuid)
        self.assertEqual(AUDIT_TMPL1['strategy_uuid'],
                         audit_template.strategy_uuid)

    def test_create(self):
        audit_template = self.mgr.create(**CREATE_AUDIT_TEMPLATE)
        expect = [
            ('POST', '/v1/audit_templates', {}, CREATE_AUDIT_TEMPLATE),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertTrue(audit_template)

    def test_delete(self):
        audit_template = self.mgr.delete(audit_template_id=AUDIT_TMPL1['uuid'])
        expect = [
            ('DELETE',
             '/v1/audit_templates/%s' % AUDIT_TMPL1['uuid'],
             {},
             None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertIsNone(audit_template)

    def test_update(self):
        patch = {'op': 'replace',
                 'value': NEW_NAME,
                 'path': '/name'}
        audit_template = self.mgr.update(audit_template_id=AUDIT_TMPL1['uuid'],
                                         patch=patch)
        expect = [
            ('PATCH',
             '/v1/audit_templates/%s' % AUDIT_TMPL1['uuid'],
             {},
             patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(NEW_NAME, audit_template.name)



#
# Copyright 2013 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from watcherclient.common import base
from watcherclient.common import utils
from watcherclient import exceptions as exc

CREATION_ATTRIBUTES = ['description', 'name', 'goal', 'strategy', 'scope']


class AuditTemplate(base.Resource):
    def __repr__(self):
        return "<AuditTemplate %s>" % self._info


class AuditTemplateManager(base.Manager):
    resource_class = AuditTemplate

    @staticmethod
    def _path(id_=None):
        return '/v1/audit_templates/%s' % id_ if id_ else '/v1/audit_templates'

    def list(self, name=None, goal=None, strategy=None, limit=None,
             sort_key=None, sort_dir=None, detail=False, marker=None):
        """Retrieve a list of audit template.

        :param name: Name of the audit template
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of audit templates to return.
            2) limit == 0, return the entire list of audit_templates.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Watcher API
               (see Watcher's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about audit_templates.

        :param marker: Optional, UUID of the last audit template of
                       the previous page.

        :returns: A list of audit templates.

        """
        if limit is not None:
            limit = int(limit)

        filters = utils.common_filters(limit, sort_key, sort_dir, marker)
        if name is not None:
            filters.append('name=%s' % name)
        if goal is not None:
            filters.append("goal=%s" % goal)
        if strategy is not None:
            filters.append("strategy=%s" % strategy)

        path = ''
        if detail:
            path += 'detail'
        if filters:
            path += '?' + '&'.join(filters)

        if limit is None:
            return self._list(self._path(path), "audit_templates")
        else:
            return self._list_pagination(self._path(path), "audit_templates",
                                         limit=limit)

    def get(self, audit_template_id):
        try:
            return self._list(self._path(audit_template_id))[0]
        except IndexError:
            return None

    def create(self, **kwargs):
        new = {}
        for (key, value) in kwargs.items():
            if key in CREATION_ATTRIBUTES:
                new[key] = value
            else:
                raise exc.InvalidAttribute()
        return self._create(self._path(), new)

    def delete(self, audit_template_id):
        return self._delete(self._path(audit_template_id))

    def update(self, audit_template_id, patch):
        return self._update(self._path(audit_template_id), patch)



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cliff.formatters import yaml_format
from osc_lib import utils
from oslo_utils import uuidutils
import six

from watcherclient._i18n import _
from watcherclient.common import command
from watcherclient.common import utils as common_utils
from watcherclient import exceptions
from watcherclient.v1 import resource_fields as res_fields


def format_global_efficacy(global_efficacy):
    formatted_global_eff = {}
    for eff in global_efficacy:
        eff_name = eff.get('name')
        if (eff.get('value') is not None and eff.get('unit')):
            formatted_global_eff[eff_name] = "%(value).2f %(unit)s" % dict(
                unit=eff.get('unit'),
                value=eff.get('value'))
        elif eff.get('value') is not None:
            formatted_global_eff[eff_name] = eff.get('value')
    return formatted_global_eff


class ShowActionPlan(command.ShowOne):
    """Show detailed information about a given action plan."""

    def get_parser(self, prog_name):
        parser = super(ShowActionPlan, self).get_parser(prog_name)
        parser.add_argument(
            'action_plan',
            metavar='<action-plan>',
            help=_('UUID of the action plan'),
        )
        return parser

    def _format_indicators(self, action_plan, parsed_args):
        out = six.StringIO()
        efficacy_indicators = action_plan.efficacy_indicators
        fields = ['name', 'description', 'value', 'unit']
        yaml_format.YAMLFormatter().emit_list(
            column_names=list(field.capitalize()
                              for field in fields),
            data=[utils.get_dict_properties(spec, fields)
                  for spec in efficacy_indicators],
            stdout=out,
            parsed_args=parsed_args,
            )
        return out.getvalue() or ''

    def _format_global_efficacy(self, global_efficacy, parsed_args):
        formatted_global_efficacy = format_global_efficacy(global_efficacy)
        out = six.StringIO()
        yaml_format.YAMLFormatter().emit_one(
            column_names=list(resource.capitalize()
                              for resource in formatted_global_efficacy),
            data=[value for value in formatted_global_efficacy.values()],
            stdout=out,
            parsed_args=parsed_args,
            )
        return out.getvalue() or ''

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        action_plan_uuid = parsed_args.action_plan

        if not uuidutils.is_uuid_like(action_plan_uuid):
            raise exceptions.ValidationError()

        try:
            action_plan = client.action_plan.get(action_plan_uuid)
        except exceptions.HTTPNotFound as exc:
            raise exceptions.CommandError(str(exc))

        if parsed_args.formatter == 'table':
            # Update the raw efficacy indicators with the formatted ones
            action_plan.efficacy_indicators = (
                self._format_indicators(action_plan, parsed_args))

            # Update the raw global efficacy with the formatted one
            action_plan.global_efficacy = self._format_global_efficacy(
                action_plan.global_efficacy, parsed_args)

        columns = res_fields.ACTION_PLAN_FIELDS
        column_headers = res_fields.ACTION_PLAN_FIELD_LABELS
        return column_headers, utils.get_item_properties(action_plan, columns)


class ListActionPlan(command.Lister):
    """List information on retrieved action plans."""

    def get_parser(self, prog_name):
        parser = super(ListActionPlan, self).get_parser(prog_name)
        parser.add_argument(
            '--audit',
            metavar='<audit>',
            help=_('UUID of an audit used for filtering.'))
        parser.add_argument(
            '--detail',
            dest='detail',
            action='store_true',
            default=False,
            help=_("Show detailed information about action plans."))
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of action plans to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Watcher API Service.'))
        parser.add_argument(
            '--marker',
            metavar='<actionplan>',
            help=_('The last actionplan UUID of the previous page; '
                   'displays list of actionplans after "marker".'))
        parser.add_argument(
            '--sort-key',
            metavar='<field>',
            help=_('Action Plan field that will be used for sorting.'))
        parser.add_argument(
            '--sort-dir',
            metavar='<direction>',
            choices=['asc', 'desc'],
            help=_('Sort direction: "asc" (the default) or "desc".'))

        return parser

    def _format_indicators(self, action_plan, parsed_args):
        out = six.StringIO()
        efficacy_indicators = action_plan.efficacy_indicators
        fields = ['name', 'value', 'unit']
        yaml_format.YAMLFormatter().emit_list(
            column_names=list(field.capitalize()
                              for field in fields),
            data=[utils.get_dict_properties(spec, fields)
                  for spec in efficacy_indicators],
            stdout=out,
            parsed_args=parsed_args,
            )
        return out.getvalue() or ''

    def _format_global_efficacy(self, global_efficacy, parsed_args):
        formatted_global_efficacy = format_global_efficacy(global_efficacy)
        out = six.StringIO()
        yaml_format.YAMLFormatter().emit_one(
            column_names=list(resource.capitalize()
                              for resource in formatted_global_efficacy),
            data=[value for value in formatted_global_efficacy.values()],
            stdout=out,
            parsed_args=parsed_args,
            )
        return out.getvalue() or ''

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        params = {}
        if parsed_args.audit is not None:
            params['audit'] = parsed_args.audit
        if parsed_args.detail:
            fields = res_fields.ACTION_PLAN_FIELDS
            field_labels = res_fields.ACTION_PLAN_FIELD_LABELS
        else:
            fields = res_fields.ACTION_PLAN_SHORT_LIST_FIELDS
            field_labels = res_fields.ACTION_PLAN_SHORT_LIST_FIELD_LABELS

        params.update(common_utils.common_params_for_list(
            parsed_args, fields, field_labels))

        data = client.action_plan.list(**params)

        if parsed_args.formatter == 'table':
            for action_plan in data:
                # Update the raw efficacy indicators with the formatted ones
                action_plan.efficacy_indicators = (
                    self._format_indicators(action_plan, parsed_args))

                # Update the raw global efficacy with the formatted one
                action_plan.global_efficacy = self._format_global_efficacy(
                    action_plan.global_efficacy, parsed_args)

        return (field_labels,
                (utils.get_item_properties(item, fields) for item in data))


class UpdateActionPlan(command.ShowOne):
    """Update action plan command."""

    def get_parser(self, prog_name):
        parser = super(UpdateActionPlan, self).get_parser(prog_name)
        parser.add_argument(
            'action_plan',
            metavar='<action-plan>',
            help=_("UUID of the action_plan."))
        parser.add_argument(
            'op',
            metavar='<op>',
            choices=['add', 'replace', 'remove'],
            help=_("Operation: 'add', 'replace', or 'remove'."))
        parser.add_argument(
            'attributes',
            metavar='<path=value>',
            nargs='+',
            action='append',
            default=[],
            help=_("Attribute to add, replace, or remove. Can be specified "
                   "multiple times. For 'remove', only <path> is necessary."))

        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        if not uuidutils.is_uuid_like(parsed_args.action_plan):
            raise exceptions.ValidationError()

        patch = common_utils.args_array_to_patch(
            parsed_args.op, parsed_args.attributes[0])

        action_plan = client.action_plan.update(parsed_args.action_plan, patch)

        columns = res_fields.ACTION_PLAN_FIELDS
        column_headers = res_fields.ACTION_PLAN_FIELD_LABELS

        return column_headers, utils.get_item_properties(action_plan, columns)


class StartActionPlan(command.ShowOne):
    """Start action plan command."""

    def get_parser(self, prog_name):
        parser = super(StartActionPlan, self).get_parser(prog_name)
        parser.add_argument(
            'action_plan',
            metavar='<action-plan>',
            help=_("UUID of the action_plan."))

        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        if not uuidutils.is_uuid_like(parsed_args.action_plan):
            raise exceptions.ValidationError()

        action_plan = client.action_plan.start(parsed_args.action_plan)

        columns = res_fields.ACTION_PLAN_FIELDS
        column_headers = res_fields.ACTION_PLAN_FIELD_LABELS

        return column_headers, utils.get_item_properties(action_plan, columns)


class DeleteActionPlan(command.Command):
    """Delete action plan command."""

    def get_parser(self, prog_name):
        parser = super(DeleteActionPlan, self).get_parser(prog_name)
        parser.add_argument(
            'action_plans',
            metavar='<action-plan>',
            nargs='+',
            help=_('UUID of the action plan'),
        )
        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        for action_plan in parsed_args.action_plans:
            if not uuidutils.is_uuid_like(action_plan):
                raise exceptions.ValidationError()

            client.action_plan.delete(action_plan)


class CancelActionPlan(command.ShowOne):
    """Cancel action plan command."""

    def get_parser(self, prog_name):
        parser = super(CancelActionPlan, self).get_parser(prog_name)
        parser.add_argument(
            'action_plan',
            metavar='<action-plan>',
            help=_("UUID of the action_plan."))

        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        if not uuidutils.is_uuid_like(parsed_args.action_plan):
            raise exceptions.ValidationError()

        action_plan = client.action_plan.cancel(parsed_args.action_plan)

        columns = res_fields.ACTION_PLAN_FIELDS
        column_headers = res_fields.ACTION_PLAN_FIELD_LABELS

        return column_headers, utils.get_item_properties(action_plan, columns)



# Copyright (c) 2016 Servionica
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest import mock

import six

from watcherclient import shell
from watcherclient.tests.unit.v1 import base
from watcherclient import v1 as resource
from watcherclient.v1 import resource_fields

SERVICE_1 = {
    'name': 'watcher-applier',
    'host': 'controller',
    'status': 'ACTIVE',
    'last_seen_up': None,
    'created_at': datetime.datetime.now().isoformat(),
    'updated_at': None,
    'deleted_at': None,
}

SERVICE_2 = {
    'name': 'watcher-decision-engine',
    'host': 'controller',
    'status': 'FAILED',
    'last_seen_up': None,
    'created_at': datetime.datetime.now().isoformat(),
    'updated_at': None,
    'deleted_at': None,
}


class ServiceShellTest(base.CommandTestCase):

    SHORT_LIST_FIELDS = resource_fields.SERVICE_SHORT_LIST_FIELDS
    SHORT_LIST_FIELD_LABELS = (
        resource_fields.SERVICE_SHORT_LIST_FIELD_LABELS)
    FIELDS = resource_fields.SERVICE_FIELDS
    FIELD_LABELS = resource_fields.SERVICE_FIELD_LABELS

    def setUp(self):
        super(self.__class__, self).setUp()

        p_service_manager = mock.patch.object(resource, 'ServiceManager')
        self.m_service_mgr_cls = p_service_manager.start()
        self.addCleanup(p_service_manager.stop)

        self.m_service_mgr = mock.Mock()
        self.m_service_mgr_cls.return_value = self.m_service_mgr

        self.stdout = six.StringIO()
        self.cmd = shell.WatcherShell(stdout=self.stdout)

    def test_do_service_list(self):
        service1 = resource.Service(mock.Mock(), SERVICE_1)
        service2 = resource.Service(mock.Mock(), SERVICE_2)
        self.m_service_mgr.list.return_value = [
            service1, service2]

        exit_code, results = self.run_cmd('service list')

        for res in results:
            del res['ID']

        self.assertEqual(0, exit_code)
        self.assertEqual(
            [self.resource_as_dict(service1, self.SHORT_LIST_FIELDS,
                                   self.SHORT_LIST_FIELD_LABELS),
             self.resource_as_dict(service2, self.SHORT_LIST_FIELDS,
                                   self.SHORT_LIST_FIELD_LABELS)],
            results)

        self.m_service_mgr.list.assert_called_once_with(detail=False)

    def test_do_service_list_detail(self):
        service1 = resource.Service(mock.Mock(), SERVICE_1)
        service2 = resource.Service(mock.Mock(), SERVICE_2)
        self.m_service_mgr.list.return_value = [
            service1, service2]

        exit_code, results = self.run_cmd('service list --detail')

        for res in results:
            del res['ID']

        self.assertEqual(0, exit_code)
        self.assertEqual(
            [self.resource_as_dict(service1, self.FIELDS,
                                   self.FIELD_LABELS),
             self.resource_as_dict(service2, self.FIELDS,
                                   self.FIELD_LABELS)],
            results)

        self.m_service_mgr.list.assert_called_once_with(detail=True)

    def test_do_service_show_by_name(self):
        service = resource.Service(mock.Mock(), SERVICE_1)
        self.m_service_mgr.get.return_value = service

        exit_code, result = self.run_cmd('service show watcher-applier')

        del result['ID']

        self.assertEqual(0, exit_code)
        self.assertEqual(
            self.resource_as_dict(service, self.FIELDS, self.FIELD_LABELS),
            result)
        self.m_service_mgr.get.assert_called_once_with('watcher-applier')



#
# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
import os
import uuid
import yaml

from oslo_serialization import jsonutils
from oslo_utils import importutils

from watcherclient._i18n import _
from watcherclient import exceptions as exc


class HelpFormatter(argparse.HelpFormatter):
    def start_section(self, heading):
        # Title-case the headings
        heading = '%s%s' % (heading[0].upper(), heading[1:])
        super(HelpFormatter, self).start_section(heading)


def define_command(subparsers, command, callback, cmd_mapper):
    '''Define a command in the subparsers collection.

    :param subparsers: subparsers collection where the command will go
    :param command: command name
    :param callback: function that will be used to process the command
    '''
    desc = callback.__doc__ or ''
    help = desc.strip().split('\n')[0]
    arguments = getattr(callback, 'arguments', [])

    subparser = subparsers.add_parser(command, help=help,
                                      description=desc,
                                      add_help=False,
                                      formatter_class=HelpFormatter)
    subparser.add_argument('-h', '--help', action='help',
                           help=argparse.SUPPRESS)
    cmd_mapper[command] = subparser
    for (args, kwargs) in arguments:
        subparser.add_argument(*args, **kwargs)
    subparser.set_defaults(func=callback)


def define_commands_from_module(subparsers, command_module, cmd_mapper):
    """Add *do_* methods in a module and add as commands into a subparsers."""

    for method_name in (a for a in dir(command_module) if a.startswith('do_')):
        # Commands should be hypen-separated instead of underscores.
        command = method_name[3:].replace('_', '-')
        callback = getattr(command_module, method_name)
        define_command(subparsers, command, callback, cmd_mapper)


def import_versioned_module(version, submodule=None):
    module = 'watcherclient.v%s' % version
    if submodule:
        module = '.'.join((module, submodule))
    return importutils.import_module(module)


def split_and_deserialize(string, exclude_fields=[]):
    """Split and try to JSON deserialize a string.

    Gets a string with the KEY=VALUE format, split it (using '=' as the
    separator) and try to JSON deserialize the VALUE.

    :returns: A tuple of (key, value).
    """
    try:
        key, value = string.split("=", 1)
    except ValueError:
        raise exc.CommandError(_('Attributes must be a list of '
                                 'PATH=VALUE not "%s"') % string)
    if key not in exclude_fields:
        try:
            value = jsonutils.loads(value)
        except ValueError:
            pass

    return (key, value)


def args_array_to_dict(kwargs, key_to_convert):
    values_to_convert = kwargs.get(key_to_convert)
    if values_to_convert:
        kwargs[key_to_convert] = dict(split_and_deserialize(v)
                                      for v in values_to_convert)
    return kwargs


def args_array_to_patch(op, attributes, exclude_fields=[]):
    patch = []
    for attr in attributes:
        # Sanitize
        if not attr.startswith('/'):
            attr = '/' + attr

        if op in ['add', 'replace']:
            path, value = split_and_deserialize(attr,
                                                exclude_fields=exclude_fields)
            patch.append({'op': op, 'path': path, 'value': value})

        elif op == "remove":
            # For remove only the key is needed
            patch.append({'op': op, 'path': attr})
        else:
            raise exc.CommandError(_('Unknown PATCH operation: %s') % op)
    return patch


def common_params_for_list(args, fields, field_labels):
    """Generate 'params' dict that is common for every 'list' command.

    :param args: arguments from command line.
    :param fields: possible fields for sorting.
    :param field_labels: possible field labels for sorting.
    :returns: a dict with params to pass to the client method.
    """
    params = {}
    if args.limit is not None:
        if args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') % args.limit)
        params['limit'] = args.limit

    if args.sort_key is not None:
        # Support using both heading and field name for sort_key
        fields_map = dict(zip(field_labels, fields))
        fields_map.update(zip(fields, fields))
        try:
            sort_key = fields_map[args.sort_key]
        except KeyError:
            raise exc.CommandError(
                _("%(sort_key)s is an invalid field for sorting, "
                  "valid values for --sort-key are: %(valid)s") %
                {'sort_key': args.sort_key,
                 'valid': list(fields_map)})
        params['sort_key'] = sort_key
    if args.sort_dir is not None:
        if args.sort_dir not in ('asc', 'desc'):
            raise exc.CommandError(
                _("%s is an invalid value for sort direction, "
                  "valid values for --sort-dir are: 'asc', 'desc'") %
                args.sort_dir)
        params['sort_dir'] = args.sort_dir

    marker = getattr(args, 'marker', None)
    if marker is not None:
        params['marker'] = marker
    params['detail'] = args.detail

    return params


def common_filters(limit=None, sort_key=None, sort_dir=None, marker=None):
    """Generate common filters for any list request.

    :param limit: maximum number of entities to return.
    :param sort_key: field to use for sorting.
    :param sort_dir: direction of sorting: 'asc' or 'desc'.
    :param marker: The last actionplan UUID of the previous page.
    :returns: list of string filters.
    """
    filters = []
    if isinstance(limit, int) and limit > 0:
        filters.append('limit=%s' % limit)
    if sort_key is not None:
        filters.append('sort_key=%s' % sort_key)
    if sort_dir is not None:
        filters.append('sort_dir=%s' % sort_dir)
    if marker is not None:
        filters.append('marker=%s' % marker)
    return filters


def is_uuid_like(val):
    """Returns validation of a value as a UUID.

    For our purposes, a UUID is a canonical form string:
    aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa

    """
    try:
        return str(uuid.UUID(val)) == val
    except (TypeError, ValueError, AttributeError):
        return False


def serialize_file_to_dict(filename):
    filename = os.path.expanduser(filename)
    with open(filename, "rb") as stream:
        scope = yaml.safe_load(stream.read())
    return scope



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

from osc_lib import utils
from oslo_utils import uuidutils

from watcherclient._i18n import _
from watcherclient.common import command
from watcherclient.common import utils as common_utils
from watcherclient import exceptions
from watcherclient.v1 import resource_fields as res_fields


class ShowAuditTemplate(command.ShowOne):
    """Show detailed information about a given audit template."""

    def get_parser(self, prog_name):
        parser = super(ShowAuditTemplate, self).get_parser(prog_name)
        parser.add_argument(
            'audit_template',
            metavar='<audit-template>',
            help=_('UUID or name of the audit template'),
        )
        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        audit_template_uuid = parsed_args.audit_template

        try:
            audit_template = client.audit_template.get(audit_template_uuid)
        except exceptions.HTTPNotFound as exc:
            raise exceptions.CommandError(str(exc))

        columns = res_fields.AUDIT_TEMPLATE_FIELDS
        column_headers = res_fields.AUDIT_TEMPLATE_FIELD_LABELS

        return column_headers, utils.get_item_properties(
            audit_template, columns)


class ListAuditTemplate(command.Lister):
    """List information on retrieved audit templates."""

    def get_parser(self, prog_name):
        parser = super(ListAuditTemplate, self).get_parser(prog_name)
        parser.add_argument(
            '--detail',
            dest='detail',
            action='store_true',
            default=False,
            help=_("Show detailed information about audit templates."))
        parser.add_argument(
            '--goal',
            dest='goal',
            metavar='<goal>',
            help=_('UUID or name of the goal used for filtering.'))
        parser.add_argument(
            '--strategy',
            dest='strategy',
            metavar='<strategy>',
            help=_('UUID or name of the strategy used for filtering.'))
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of audit templates to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Watcher API Service.'))
        parser.add_argument(
            '--sort-key',
            metavar='<field>',
            help=_('Audit template field that will be used for sorting.'))
        parser.add_argument(
            '--sort-dir',
            metavar='<direction>',
            choices=['asc', 'desc'],
            help=_('Sort direction: "asc" (the default) or "desc".'))
        parser.add_argument(
            '--marker',
            dest='marker',
            metavar='<marker>',
            default=None,
            help=_('UUID of the last audit template of the previous page; '
                   'displays list of audit templates after "marker".'))

        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        params = {}

        # Optional
        if parsed_args.goal:
            params['goal'] = parsed_args.goal

        # Optional
        if parsed_args.strategy:
            params['strategy'] = parsed_args.strategy

        if parsed_args.detail:
            fields = res_fields.AUDIT_TEMPLATE_FIELDS
            field_labels = res_fields.AUDIT_TEMPLATE_FIELD_LABELS
        else:
            fields = res_fields.AUDIT_TEMPLATE_SHORT_LIST_FIELDS
            field_labels = res_fields.AUDIT_TEMPLATE_SHORT_LIST_FIELD_LABELS

        params.update(common_utils.common_params_for_list(
            parsed_args, fields, field_labels))

        data = client.audit_template.list(**params)

        return (field_labels,
                (utils.get_item_properties(item, fields) for item in data))


class CreateAuditTemplate(command.ShowOne):
    """Create new audit template."""

    def get_parser(self, prog_name):
        class SmartFormatter(argparse.HelpFormatter):

            def _split_lines(self, text, width):
                if '\n' in text:
                    return text.splitlines()
                else:
                    return argparse.HelpFormatter._split_lines(
                        self, text, width)

        parser = super(CreateAuditTemplate, self).get_parser(
            prog_name, formatter_class=SmartFormatter)
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Name for this audit template.'))
        parser.add_argument(
            'goal',
            metavar='<goal>',
            help=_('Goal UUID or name associated to this audit template.'))
        parser.add_argument(
            '-s', '--strategy',
            dest='strategy',
            metavar='<strategy>',
            help=_('Strategy UUID or name associated to this audit template.'))
        parser.add_argument(
            '-d', '--description',
            metavar='<description>',
            help=_('Description of the audit template.'))
        parser.add_argument(
            '--scope',
            metavar='<path>',
            help=_("Part of the cluster on which an audit will be done.\n"
                   "Can be provided either in yaml or json file.\n"
                   "YAML example::\n"
                   "\n"
                   " - compute:\n"
                   "   - host_aggregates:\n"
                   "     - id: 1\n"
                   "     - id: 2\n"
                   "     - id: 3\n"
                   "   - availability_zones:\n"
                   "     - name: AZ1\n"
                   "     - name: AZ2\n"
                   "   - exclude:\n"
                   "     - instances:\n"
                   "       - uuid: UUID1\n"
                   "       - uuid: UUID2\n"
                   "     - compute_nodes:\n"
                   "       - name: compute1\n"
                   " - storage: \n"
                   "   - availability_zones:\n"
                   "     - name: AZ1\n"
                   "     - name: AZ2\n"
                   "   - volume_types:\n"
                   "     - name: lvm1\n"
                   "     - name: lvm2\n"
                   "   - exclude:\n"
                   "     - storage_pools:\n"
                   "       - name: host0@backend0#pool0\n"
                   "       - name: host1@backend1#pool1\n"
                   "     - volumes:\n"
                   "       - uuid: UUID1\n"
                   "       - uuid: UUID2\n"
                   "     - projects:\n"
                   "       - uuid: UUID1\n"
                   "       - uuid: UUID2\n"
                   "\n"
                   "JSON example::\n"
                   "\n"
                   " [\n"
                   "  {\"compute\":\n"
                   "       [{\"host_aggregates\": [\n"
                   "             {\"id\": 1},\n"
                   "             {\"id\": 2},\n"
                   "             {\"id\": 3}]},\n"
                   "        {\"availability_zones\": [\n"
                   "             {\"name\": \"AZ1\"},\n"
                   "             {\"name\": \"AZ2\"}]},\n"
                   "        {\"exclude\": [\n"
                   "             {\"instances\": [\n"
                   "                  {\"uuid\": \"UUID1\"},\n"
                   "                  {\"uuid\": \"UUID2\"}\n"
                   "             ]},\n"
                   "             {\"compute_nodes\": [\n"
                   "                  {\"name\": \"compute1\"}\n"
                   "             ]}\n"
                   "        ]}]\n"
                   "   },\n"
                   "  {\"storage\":\n"
                   "       [{\"availability_zones\": [\n"
                   "             {\"name\": \"AZ1\"},\n"
                   "             {\"name\": \"AZ2\"}]},\n"
                   "        {\"volume_types\": [\n"
                   "             {\"name\": \"lvm1\"},\n"
                   "             {\"name\": \"lvm2\"}]},\n"
                   "        {\"exclude\": [\n"
                   "             {\"storage_pools\": [\n"
                   "                  {\"name\": \"host0@backend0#pool0\"},\n"
                   "                  {\"name\": \"host1@backend1#pool1\"}\n"
                   "             ]},\n"
                   "             {\"volumes\": [\n"
                   "                  {\"uuid\": \"UUID1\"},\n"
                   "                  {\"uuid\": \"UUID2\"}\n"
                   "             ]},\n"
                   "             {\"projects\": [\n"
                   "                  {\"uuid\": \"UUID1\"},\n"
                   "                  {\"uuid\": \"UUID2\"}\n"
                   "             ]},\n"
                   "        ]}]\n"
                   "   }\n"
                   " ]\n"
                   )
        )

        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        field_list = ['description', 'name', 'goal', 'strategy', 'scope']
        fields = dict((k, v) for (k, v) in vars(parsed_args).items()
                      if k in field_list and v is not None)

        # mandatory
        if not uuidutils.is_uuid_like(fields['goal']):
            fields['goal'] = client.goal.get(fields['goal']).uuid

        # optional
        if fields.get('strategy'):
            if not uuidutils.is_uuid_like(fields['strategy']):
                fields['strategy'] = client.strategy.get(
                    fields['strategy']).uuid
        if fields.get('scope'):
            fields['scope'] = common_utils.serialize_file_to_dict(
                fields['scope'])

        audit_template = client.audit_template.create(**fields)

        columns = res_fields.AUDIT_TEMPLATE_FIELDS
        column_headers = res_fields.AUDIT_TEMPLATE_FIELD_LABELS

        return (column_headers,
                utils.get_item_properties(audit_template, columns))


class UpdateAuditTemplate(command.ShowOne):
    """Update audit template command."""

    def get_parser(self, prog_name):
        parser = super(UpdateAuditTemplate, self).get_parser(prog_name)
        parser.add_argument(
            'audit_template',
            metavar='<audit-template>',
            help=_("UUID or name of the audit_template."))
        parser.add_argument(
            'op',
            metavar='<op>',
            choices=['add', 'replace', 'remove'],
            help=_("Operation: 'add', 'replace', or 'remove'."))
        parser.add_argument(
            'attributes',
            metavar='<path=value>',
            nargs='+',
            action='append',
            default=[],
            help=_("Attribute to add, replace, or remove. Can be specified "
                   "multiple times. For 'remove', only <path> is necessary."))

        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        patch = common_utils.args_array_to_patch(
            parsed_args.op, parsed_args.attributes[0])

        audit_template = client.audit_template.update(
            parsed_args.audit_template, patch)

        columns = res_fields.AUDIT_TEMPLATE_FIELDS
        column_headers = res_fields.AUDIT_TEMPLATE_FIELD_LABELS

        return column_headers, utils.get_item_properties(
            audit_template, columns)


class DeleteAuditTemplate(command.Command):
    """Delete audit template command."""

    def get_parser(self, prog_name):
        parser = super(DeleteAuditTemplate, self).get_parser(prog_name)
        parser.add_argument(
            'audit_templates',
            metavar='<audit-template>',
            nargs='+',
            help=_('UUID or name of the audit template'),
        )
        return parser

    def take_action(self, parsed_args):
        client = getattr(self.app.client_manager, "infra-optim")

        for audit_template in parsed_args.audit_templates:
            client.audit_template.delete(audit_template)



# Copyright (c) 2016 b<>com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest import mock

from oslo_utils.uuidutils import generate_uuid
import six

from watcherclient import exceptions
from watcherclient import shell
from watcherclient.tests.unit.v1 import base
from watcherclient import v1 as resource
from watcherclient.v1 import resource_fields

ACTION_PLAN_1 = {
    'uuid': 'd9d9978e-6db5-4a05-8eab-1531795d7004',
    'audit_uuid': '770ef053-ecb3-48b0-85b5-d55a2dbc6588',
    'strategy_name': 'dummy',
    'state': 'RECOMMENDED',
    'efficacy_indicators': [{'description': 'Indicator 1',
                             'name': 'indicator1',
                             'unit': '%'}],
    'created_at': datetime.datetime.now().isoformat(),
    'updated_at': None,
    'global_efficacy': [
        {"value": 99,
         "unit": "%",
         "name": "dummy_global_efficacy",
         "description": "Dummy Global Efficacy"},
        {"value": 75,
         "unit": "%",
         "name": "dummy_global_efficacy2",
         "description": "Dummy Global Efficacy2"}
        ],
    'deleted_at': None,
    'hostname': ''
}

ACTION_PLAN_2 = {
    'uuid': 'd6363285-5afa-4a26-96f2-89441e335765',
    'audit_uuid': '239f02a5-9649-4e14-9d33-ac2bf67cb755',
    'strategy_name': 'dummy',
    'state': 'RECOMMENDED',
    'created_at': datetime.datetime.now().isoformat(),
    'efficacy_indicators': [{'description': 'Indicator 2',
                             'name': 'indicator2',
                             'unit': '%'}],
    'updated_at': None,
    'global_efficacy': [{
        "value": 87,
        "unit": "%",
        "name": "dummy_global_efficacy",
        "description": "Dummy Global Efficacy",
    }],
    'deleted_at': None,
    'hostname': ''
}


class ActionPlanShellTest(base.CommandTestCase):

    SHORT_LIST_FIELDS = resource_fields.ACTION_PLAN_SHORT_LIST_FIELDS
    SHORT_LIST_FIELD_LABELS = (
        resource_fields.ACTION_PLAN_SHORT_LIST_FIELD_LABELS)
    FIELDS = resource_fields.ACTION_PLAN_FIELDS
    FIELD_LABELS = resource_fields.ACTION_PLAN_FIELD_LABELS
    GLOBAL_EFFICACY_FIELDS = resource_fields.GLOBAL_EFFICACY_FIELDS

    def setUp(self):
        super(self.__class__, self).setUp()

        p_audit_manager = mock.patch.object(resource, 'AuditManager')
        p_audit_template_manager = mock.patch.object(
            resource, 'ActionPlanManager')
        p_action_plan_manager = mock.patch.object(
            resource, 'ActionPlanManager')

        self.m_audit_mgr_cls = p_audit_manager.start()
        self.m_audit_template_mgr_cls = p_audit_template_manager.start()
        self.m_action_plan_mgr_cls = p_action_plan_manager.start()

        self.addCleanup(p_audit_manager.stop)
        self.addCleanup(p_audit_template_manager.stop)
        self.addCleanup(p_action_plan_manager.stop)

        self.m_audit_mgr = mock.Mock()
        self.m_audit_template_mgr = mock.Mock()
        self.m_action_plan_mgr = mock.Mock()

        self.m_audit_mgr_cls.return_value = self.m_audit_mgr
        self.m_audit_template_mgr_cls.return_value = self.m_audit_template_mgr
        self.m_action_plan_mgr_cls.return_value = self.m_action_plan_mgr

        self.stdout = six.StringIO()
        self.cmd = shell.WatcherShell(stdout=self.stdout)

    def test_do_action_plan_list(self):
        action_plan1 = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        action_plan2 = resource.ActionPlan(mock.Mock(), ACTION_PLAN_2)
        self.m_action_plan_mgr.list.return_value = [
            action_plan1, action_plan2]

        exit_code, results = self.run_cmd('actionplan list')

        self.assertEqual(0, exit_code)
        self.assertEqual(
            [self.resource_as_dict(action_plan1, self.SHORT_LIST_FIELDS,
                                   self.SHORT_LIST_FIELD_LABELS),
             self.resource_as_dict(action_plan2, self.SHORT_LIST_FIELDS,
                                   self.SHORT_LIST_FIELD_LABELS)],
            results)

        self.assertEqual(action_plan1.global_efficacy,
                         results[0]['Global efficacy'])
        self.assertEqual(action_plan2.global_efficacy,
                         results[1]['Global efficacy'])

    def test_do_action_plan_list_by_table(self):
        action_plan1 = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        action_plan2 = resource.ActionPlan(mock.Mock(), ACTION_PLAN_2)
        self.m_action_plan_mgr.list.return_value = [
            action_plan1, action_plan2]

        exit_code, results = self.run_cmd('actionplan list', 'table')
        self.assertEqual(0, exit_code)
        self.assertIn(ACTION_PLAN_1['uuid'], results)
        self.assertIn(ACTION_PLAN_2['uuid'], results)

        self.m_action_plan_mgr.list.assert_called_once_with(detail=False)

    def test_do_action_plan_list_detail(self):
        action_plan1 = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        action_plan2 = resource.ActionPlan(mock.Mock(), ACTION_PLAN_2)
        self.m_action_plan_mgr.list.return_value = [
            action_plan1, action_plan2]

        exit_code, results = self.run_cmd('actionplan list --detail')

        self.assertEqual(0, exit_code)
        self.assertEqual(
            [self.resource_as_dict(action_plan1, self.FIELDS,
                                   self.FIELD_LABELS),
             self.resource_as_dict(action_plan2, self.FIELDS,
                                   self.FIELD_LABELS)],
            results)
        self.assertEqual(action_plan1.global_efficacy,
                         results[0]['Global efficacy'])
        self.assertEqual(action_plan2.global_efficacy,
                         results[1]['Global efficacy'])

        self.m_action_plan_mgr.list.assert_called_once_with(detail=True)

    def test_do_action_plan_list_filter_by_audit(self):
        action_plan1 = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        self.m_action_plan_mgr.list.return_value = [action_plan1]

        exit_code, results = self.run_cmd(
            'actionplan list --audit '
            '770ef053-ecb3-48b0-85b5-d55a2dbc6588')

        self.assertEqual(0, exit_code)
        self.assertEqual(
            [self.resource_as_dict(action_plan1, self.SHORT_LIST_FIELDS,
                                   self.SHORT_LIST_FIELD_LABELS)],
            results)

        self.m_action_plan_mgr.list.assert_called_once_with(
            detail=False,
            audit='770ef053-ecb3-48b0-85b5-d55a2dbc6588',
        )

    def test_do_action_plan_show_by_uuid(self):
        action_plan = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        self.m_action_plan_mgr.get.return_value = action_plan

        exit_code, result = self.run_cmd(
            'actionplan show d9d9978e-6db5-4a05-8eab-1531795d7004')

        self.assertEqual(0, exit_code)
        self.assertEqual(
            self.resource_as_dict(
                action_plan, self.FIELDS, self.FIELD_LABELS),
            result)
        self.assertEqual(action_plan.global_efficacy,
                         result['Global efficacy'])
        self.m_action_plan_mgr.get.assert_called_once_with(
            'd9d9978e-6db5-4a05-8eab-1531795d7004')

    def test_do_action_plan_show_by_not_uuid(self):
        self.m_action_plan_mgr.get.side_effect = exceptions.HTTPNotFound

        exit_code, result = self.run_cmd(
            'actionplan show not_uuid', formatting=None)

        self.assertEqual(1, exit_code)
        self.assertEqual('', result)

    def test_do_action_plan_show_by_random_uuid(self):
        # verify that show a wrong actionplan will raise Exception
        self.m_action_plan_mgr.get.side_effect = exceptions.HTTPNotFound
        fake_uuid = generate_uuid()

        exit_code, result = self.run_cmd(
            'actionplan show {}'.format(fake_uuid), formatting=None)
        self.assertEqual(1, exit_code)
        self.assertEqual('', result)

        self.m_action_plan_mgr.get.assert_called_once_with(fake_uuid)

    def test_do_action_plan_show_uuid_by_table(self):
        # verify that show an actionplan can be in a 'table' format
        action_plan = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        self.m_action_plan_mgr.get.return_value = action_plan

        exit_code, result = self.run_cmd(
            'actionplan show d9d9978e-6db5-4a05-8eab-1531795d7004',
            formatting='table')
        self.assertEqual(0, exit_code)
        self.assertIn(ACTION_PLAN_1['uuid'], result)

        self.m_action_plan_mgr.get.assert_called_once_with(
            'd9d9978e-6db5-4a05-8eab-1531795d7004')

    def test_do_action_plan_delete(self):
        self.m_action_plan_mgr.delete.return_value = ''

        exit_code, result = self.run_cmd(
            'actionplan delete 5869da81-4876-4687-a1ed-12cd64cf53d9',
            formatting=None)

        self.assertEqual(0, exit_code)
        self.assertEqual('', result)
        self.m_action_plan_mgr.delete.assert_called_once_with(
            '5869da81-4876-4687-a1ed-12cd64cf53d9')

    def test_do_action_plan_delete_not_uuid(self):
        exit_code, result = self.run_cmd(
            'actionplan delete not_uuid', formatting=None)

        self.assertEqual(1, exit_code)
        self.assertEqual('', result)

    def test_do_action_plan_delete_multiple(self):
        self.m_action_plan_mgr.delete.return_value = ''

        exit_code, result = self.run_cmd(
            'actionplan delete 5869da81-4876-4687-a1ed-12cd64cf53d9 '
            'c20627fa-ea70-4d56-ae15-4106358f773b',
            formatting=None)

        self.assertEqual(0, exit_code)
        self.assertEqual('', result)
        self.m_action_plan_mgr.delete.assert_any_call(
            '5869da81-4876-4687-a1ed-12cd64cf53d9')
        self.m_action_plan_mgr.delete.assert_any_call(
            'c20627fa-ea70-4d56-ae15-4106358f773b')

    def test_do_action_plan_update(self):
        action_plan = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        self.m_action_plan_mgr.update.return_value = action_plan

        exit_code, result = self.run_cmd(
            'actionplan update 5869da81-4876-4687-a1ed-12cd64cf53d9 '
            'replace state=CANCELLED')

        self.assertEqual(0, exit_code)
        self.assertEqual(
            self.resource_as_dict(action_plan, self.FIELDS, self.FIELD_LABELS),
            result)
        self.m_action_plan_mgr.update.assert_called_once_with(
            '5869da81-4876-4687-a1ed-12cd64cf53d9',
            [{'op': 'replace', 'path': '/state', 'value': 'CANCELLED'}])

    def test_do_action_plan_update_not_uuid(self):
        exit_code, result = self.run_cmd(
            'actionplan update not_uuid '
            'replace state=CANCELLED',
            formatting=None)

        self.assertEqual(1, exit_code)
        self.assertEqual('', result)

    def test_do_action_plan_start(self):
        action_plan = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        self.m_action_plan_mgr.start.return_value = action_plan

        exit_code, result = self.run_cmd(
            'actionplan start 5869da81-4876-4687-a1ed-12cd64cf53d9')

        self.assertEqual(0, exit_code)
        self.assertEqual(
            self.resource_as_dict(action_plan, self.FIELDS, self.FIELD_LABELS),
            result)
        self.m_action_plan_mgr.start.assert_called_once_with(
            '5869da81-4876-4687-a1ed-12cd64cf53d9')

    def test_do_action_plan_start_not_uuid(self):
        exit_code, result = self.run_cmd(
            'actionplan start not_uuid',
            formatting=None)

        self.assertEqual(1, exit_code)
        self.assertEqual('', result)

    def test_do_action_plan_cancel(self):
        action_plan = resource.ActionPlan(mock.Mock(), ACTION_PLAN_1)
        self.m_action_plan_mgr.cancel.return_value = action_plan

        exit_code, result = self.run_cmd(
            'actionplan cancel 5869da81-4876-4687-a1ed-12cd64cf53d9')

        self.assertEqual(0, exit_code)
        self.assertEqual(
            self.resource_as_dict(action_plan, self.FIELDS, self.FIELD_LABELS),
            result)
        self.m_action_plan_mgr.cancel.assert_called_once_with(
            '5869da81-4876-4687-a1ed-12cd64cf53d9')

    def test_do_action_plan_cancel_not_uuid(self):
        exit_code, result = self.run_cmd(
            'actionplan cancel not_uuid',
            formatting=None)

        self.assertEqual(1, exit_code)
        self.assertEqual('', result)



# Copyright 2016 Mirantis
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from watcherclient.common import api_versioning
from watcherclient import exceptions
from watcherclient.tests.unit import utils


class APIVersionTestCase(utils.BaseTestCase):
    def test_valid_version_strings(self):
        def _test_string(version, exp_major, exp_minor):
            v = api_versioning.APIVersion(version)
            self.assertEqual(v.ver_major, exp_major)
            self.assertEqual(v.ver_minor, exp_minor)

        _test_string("1.1", 1, 1)
        _test_string("2.10", 2, 10)
        _test_string("5.234", 5, 234)
        _test_string("12.5", 12, 5)
        _test_string("2.0", 2, 0)
        _test_string("2.200", 2, 200)

    def test_null_version(self):
        v = api_versioning.APIVersion()
        self.assertTrue(v.is_null())

    def test_invalid_version_strings(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "2")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "200")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "2.1.4")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "200.23.66.3")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "5 .3")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "5. 3")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "5.03")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "02.1")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "2.001")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, " 2.1")

        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.APIVersion, "2.1 ")

    def test_version_comparisons(self):
        v1 = api_versioning.APIVersion("2.0")
        v2 = api_versioning.APIVersion("2.5")
        v3 = api_versioning.APIVersion("5.23")
        v4 = api_versioning.APIVersion("2.0")
        v_null = api_versioning.APIVersion()

        self.assertLess(v1, v2)
        self.assertGreater(v3, v2)
        self.assertNotEqual(v1, v2)
        self.assertEqual(v1, v4)
        self.assertNotEqual(v1, v_null)
        self.assertEqual(v_null, v_null)
        self.assertRaises(TypeError, v1.__le__, "2.1")

    def test_version_matches(self):
        v1 = api_versioning.APIVersion("2.0")
        v2 = api_versioning.APIVersion("2.5")
        v3 = api_versioning.APIVersion("2.45")
        v4 = api_versioning.APIVersion("3.3")
        v5 = api_versioning.APIVersion("3.23")
        v6 = api_versioning.APIVersion("2.0")
        v7 = api_versioning.APIVersion("3.3")
        v8 = api_versioning.APIVersion("4.0")
        v_null = api_versioning.APIVersion()

        self.assertTrue(v2.matches(v1, v3))
        self.assertTrue(v2.matches(v1, v_null))
        self.assertTrue(v1.matches(v6, v2))
        self.assertTrue(v4.matches(v2, v7))
        self.assertTrue(v4.matches(v_null, v7))
        self.assertTrue(v4.matches(v_null, v8))
        self.assertFalse(v1.matches(v2, v3))
        self.assertFalse(v5.matches(v2, v4))
        self.assertFalse(v2.matches(v3, v1))

        self.assertRaises(ValueError, v_null.matches, v1, v3)

    def test_get_string(self):
        v1_string = "3.23"
        v1 = api_versioning.APIVersion(v1_string)
        self.assertEqual(v1_string, v1.get_string())

        self.assertRaises(ValueError,
                          api_versioning.APIVersion().get_string)


class GetAPIVersionTestCase(utils.BaseTestCase):
    def test_get_available_client_versions(self):
        output = api_versioning.get_available_major_versions()
        self.assertNotEqual([], output)

    def test_wrong_format(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.get_api_version, "something_wrong")

    def test_wrong_major_version(self):
        self.assertRaises(exceptions.UnsupportedVersion,
                          api_versioning.get_api_version, "2")

    @mock.patch("watcherclient.common.api_versioning.APIVersion")
    def test_only_major_part_is_presented(self, mock_apiversion):
        version = 7
        self.assertEqual(mock_apiversion.return_value,
                         api_versioning.get_api_version(version))
        mock_apiversion.assert_called_once_with("%s.0" % str(version))

    @mock.patch("watcherclient.common.api_versioning.APIVersion")
    def test_major_and_minor_parts_is_presented(self, mock_apiversion):
        version = "2.7"
        self.assertEqual(mock_apiversion.return_value,
                         api_versioning.get_api_version(version))
        mock_apiversion.assert_called_once_with(version)
